/** 
 *
 * XMLSec library
 * 
 * RSA Algorithm support
 * 
 * See Copyright for the status of this software.
 * 
 * Author: Aleksey Sanin <aleksey@aleksey.com>
 */
#include <stdlib.h>
#include <string.h>

#include <openssl/rsa.h>
#include <openssl/sha.h>
#include <openssl/objects.h>

#include <libxml/tree.h>

#include <xmlsec/xmlsec.h>
#include <xmlsec/xmltree.h>
#include <xmlsec/bn.h>
#include <xmlsec/transforms.h>
#include <xmlsec/digests.h>
#include <xmlsec/keys.h>
#include <xmlsec/base64.h>
#include <xmlsec/rsa.h>



/**
 * RSA transform
 */
static xmlSecTransformPtr xmlSecSignRsaSha1Create(xmlSecTransformId id);
static void 	xmlSecSignRsaSha1Destroy	(xmlSecTransformPtr transform);
static int  	xmlSecSignRsaSha1AddKey		(xmlSecBinTransformPtr transform, 
						 xmlSecKeyPtr key);
static int 	xmlSecSignRsaSha1Update		(xmlSecDigestTransformPtr digest,
						 const unsigned char *buffer,
						 size_t size);
static int 	xmlSecSignRsaSha1Sign		(xmlSecDigestTransformPtr digest,
						 unsigned char **buffer,
						 size_t *size);
static int 	xmlSecSignRsaSha1Verify		(xmlSecDigestTransformPtr digest,
						 const unsigned char *buffer,
						 size_t size);

/**
 * RSA key
 */
static xmlSecKeyPtr	xmlSecSignRsaKeyCreate	(xmlSecKeyId id);
static void		xmlSecSignRsaKeyDestroy	(xmlSecKeyPtr key);
static int		xmlSecSignRsaKeyGenerate	(xmlSecKeyPtr key,
							 void *context);
static int		xmlSecSignRsaKeyRead	(xmlSecKeyPtr key,
							 xmlNodePtr node);
static int		xmlSecSignRsaKeyWrite	(xmlSecKeyPtr key,
							 xmlSecKeyType type,
							 xmlNodePtr parent);

struct _xmlSecKeyId xmlSecSignRsaKeyId = {
    /* xlmlSecKeyId data  */
    BAD_CAST "RSAKeyValue",		/* const xmlChar *keyValueNodeName; */
    xmlSecDSigNs, 			/* const xmlChar *keyValueNodeNs; */
    
    /* xmlSecKeyId methods */
    xmlSecSignRsaKeyGenerate,	/* xmlSecKeyGenerateMethod generate; */
    xmlSecSignRsaKeyCreate,		/* xmlSecKeyCreateMethod create; */    
    xmlSecSignRsaKeyDestroy,	/* xmlSecKeyDestroyMethod destroy; */
    xmlSecSignRsaKeyRead, 		/* xmlSecKeyReadXmlMethod read; */
    xmlSecSignRsaKeyWrite		/* xmlSecKeyWriteXmlMethod write; */
};
xmlSecKeyId xmlSecSignRsaKey = &xmlSecSignRsaKeyId;



struct _xmlSecDigestTransformId xmlSecSignRsaSha1Id = {
    /* same as xmlSecTransformId */    
    xmlSecTransformTypeBinary,		/* xmlSecTransformType type; */
    xmlSecUsageDSigSignature,		/* xmlSecTransformUsage usage; */
    BAD_CAST "http://www.w3.org/2000/09/xmldsig#rsa-sha1", /* xmlChar *href; */
    
    xmlSecSignRsaSha1Create,		/* xmlSecTransformCreateMethod create; */
    xmlSecSignRsaSha1Destroy,		/* xmlSecTransformDestroyMethod destroy; */
    NULL,				/* xmlSecTransformReadNodeMethod read; */
    NULL,				/* xmlSecTransformWriteNodeMethod write; */
    
    /* xmlSecBinTransform data/methods */
    &xmlSecSignRsaKeyId,
    xmlSecBinTransformSubTypeDigest,	/* xmlSecBinTransformSubType binSubType; */
            
    xmlSecSignRsaSha1AddKey,		/* xmlSecBinTransformAddKeyMethod addBinKey; */
    xmlSecDigestTransformRead,		/* xmlSecBinTransformReadMethod readBin; */
    xmlSecDigestTransformWrite,		/* xmlSecBinTransformWriteMethod writeBin; */
    xmlSecDigestTransformFlush,		/* xmlSecBinTransformFlushMethod flushBin; */
    
    /* xmlSecDigestTransform data/methods */
    xmlSecSignRsaSha1Update,		/* xmlSecDigestUpdateMethod digestUpdate; */
    xmlSecSignRsaSha1Sign,		/* xmlSecDigestSignMethod digestSign; */
    xmlSecSignRsaSha1Verify		/* xmlSecDigestVerifyMethod digestVerify; */
};
xmlSecTransformId xmlSecSignRsaSha1 = (xmlSecTransformId)&xmlSecSignRsaSha1Id;


#define xmlSecSignRsaKey( k ) 			((RSA*)(( k )->keyData))

/**
 * RSA transform
 */
#define XMLSEC_RSASHA1_TRANSFORM_SIZE \
    (sizeof(xmlSecDigestTransform) + sizeof(SHA_CTX))
#define xmlSecSignRsaSha1Context(t) \
    ((SHA_CTX*)(((xmlSecDigestTransformPtr)( t ))->digestData))
#define xmlSecSignRsaSha1ContextRsa(t) \
    ((RSA*)(((xmlSecDigestTransformPtr)( t ))->binData))

/**
 * xmlSecSignRsaSha1Create:
 * @id:
 *
 *
 *
 */
static xmlSecTransformPtr 
xmlSecSignRsaSha1Create(xmlSecTransformId id) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecSignRsaSha1Create";
    xmlSecDigestTransformPtr digest;
    
    if(id != xmlSecSignRsaSha1){
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: id is not recognized\n",
	    func);
#endif 	    
	return(NULL);
    }

    /*
     * Allocate a new xmlSecBinTransform and fill the fields.
     */
    digest = (xmlSecDigestTransformPtr) xmlMalloc(XMLSEC_RSASHA1_TRANSFORM_SIZE);
    if(digest == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: XMLSEC_RSASHA1_TRANSFORM_SIZE malloc failed\n",
	    func);	
#endif 	    
	return(NULL);
    }
    memset(digest, 0, XMLSEC_RSASHA1_TRANSFORM_SIZE);
    
    digest->id = (xmlSecDigestTransformId)id;
    digest->digestData = ((unsigned char*)digest) + sizeof(xmlSecDigestTransform);

    SHA1_Init(xmlSecSignRsaSha1Context(digest)); 
    return((xmlSecTransformPtr)digest);
}

/**
 * xmlSecSignRsaSha1Destroy
 * @transform
 * 
 *
 *
 */
static void 
xmlSecSignRsaSha1Destroy(xmlSecTransformPtr transform) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecSignRsaSha1Destroy";
    xmlSecDigestTransformPtr digest;
    
    if(!xmlSecTransformCheckId(transform, xmlSecSignRsaSha1)) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: transform is invalid\n",
	    func);	
#endif 	    
	return;
    }    
    digest = (xmlSecDigestTransformPtr)transform;

    if(xmlSecSignRsaSha1ContextRsa(transform) != NULL) {
	RSA_free(xmlSecSignRsaSha1ContextRsa(transform));
    }
    
    if(digest->digest != NULL) {
	memset(digest->digest, 0, digest->digestSize);
	xmlFree(digest->digest);
    }    
        
    memset(digest, 0, XMLSEC_RSASHA1_TRANSFORM_SIZE);
    xmlFree(digest);
}

/**
 * xmlSecSignRsaSha1Update
 * @digest:
 * @buffer:
 * @size:
 *
 */
static int
xmlSecSignRsaSha1Update(xmlSecDigestTransformPtr digest,
			const unsigned char *buffer, size_t size) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecSignRsaSha1Update";
    
    if(!xmlSecTransformCheckId(digest, xmlSecSignRsaSha1)) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: transform is invalid\n",
	    func);	
#endif 	    
	return(-1);
    }    
    
    if((buffer == NULL) || (size == 0) || (digest->status != xmlSecTransformStatusNone)) {
	/* nothing to update */
	return(0);
    }
    
    SHA1_Update(xmlSecSignRsaSha1Context(digest), buffer, size); 
    return(0);
}

/**
 * xmlSecSignRsaSha1Sign
 * @digest:
 * @buffer:
 * @size
 *
 */
static int
xmlSecSignRsaSha1Sign(xmlSecDigestTransformPtr digest,
			unsigned char **buffer, size_t *size) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecSignRsaSha1Sign";
    unsigned char buf[SHA_DIGEST_LENGTH]; 
    int ret;
        
    if(!xmlSecTransformCheckId(digest, xmlSecSignRsaSha1) || 
      (xmlSecSignRsaSha1ContextRsa(digest) == NULL) ||
      ((xmlSecSignRsaSha1ContextRsa(digest)->d) == NULL) ||
      (digest->digest == NULL)) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: digest is invalid or the RSA key is null or not private\n",
	    func);	
#endif 	    
	return(-1);
    }    
    if(digest->status != xmlSecTransformStatusNone) {
	return(0);
    }    
    SHA1_Final(buf, xmlSecSignRsaSha1Context(digest)); 
    
    ret = RSA_sign(NID_sha1, buf, SHA_DIGEST_LENGTH, 
		digest->digest, &(digest->digestSize), 
		xmlSecSignRsaSha1ContextRsa(digest));
    if(ret != 1) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
    	    "%s: RSA sign failed\n",
	    func);
#endif 	    
	return(-1);	    
    }
    
    if(buffer != NULL) {
	(*buffer) = digest->digest;
    }        
    if(size != NULL) {
	(*size) = digest->digestSize;
    }        
    digest->status = xmlSecTransformStatusOk;
    return(0);
}

/**
 * xmlSecSignRsaSha1Verify
 * @digest:
 * @buffer:
 * @size:
 *
 */
static int
xmlSecSignRsaSha1Verify(xmlSecDigestTransformPtr digest,
			const unsigned char *buffer, size_t size) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecSignRsaSha1Verify";
    unsigned char buf[SHA_DIGEST_LENGTH]; 
    int ret;
        
    if(!xmlSecTransformCheckId(digest, xmlSecSignRsaSha1) ||
       (xmlSecSignRsaSha1ContextRsa(digest) == NULL)) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: digest is invalid or rsa key is null\n",
	    func);	
#endif 	    
	return(-1);
    }    
    if(buf == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: buffer is null or has an invalid size (%d)\n",
	    func, size);	
#endif 	    
	digest->status = xmlSecTransformStatusFail;
	return(-1);
    }

    SHA1_Final(buf, xmlSecSignRsaSha1Context(digest)); 
    
    ret = RSA_verify(NID_sha1, buf, SHA_DIGEST_LENGTH, 
		     (unsigned char *)buffer, size, 
		     xmlSecSignRsaSha1ContextRsa(digest));
    if(ret == 1) {
	digest->status = xmlSecTransformStatusOk;
    } else if(ret == 0) {
	digest->status = xmlSecTransformStatusFail;
    } else {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: RSA_do_verify failed\n",
	    func);
#endif 	    
	return(-1);
    }
    
    return(0);
}

/**
 * xmlSecSignRsaSha1AddKey:
 * @transform:
 * @key:
 *
 */																 
static int
xmlSecSignRsaSha1AddKey	(xmlSecBinTransformPtr transform, xmlSecKeyPtr key) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecSignRsaSha1AddKey";
    xmlSecDigestTransformPtr digest;
    RSA *rsa;
    void *digestBuf;
    
    if(!xmlSecTransformCheckId(transform, xmlSecSignRsaSha1) || 
       !xmlSecKeyCheckId(key, xmlSecSignRsaKey)) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: transform or key is invalid\n",
	    func);	
#endif 	    
	return(-1);
    }    
    digest = (xmlSecDigestTransformPtr)transform;

    if(xmlSecSignRsaKey(key) == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: key rsa data is null\n",
	    func);	
#endif 	    
	return(-1);
    }

    rsa = RSA_new();
    if(rsa == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create rsa key\n",
	    func);	
#endif 	    
	return(-1);
    }

    if(xmlSecSignRsaKey(key)->n != NULL) {
	rsa->n = BN_dup(xmlSecSignRsaKey(key)->n);
    }
    if(xmlSecSignRsaKey(key)->e != NULL) {
	rsa->e = BN_dup(xmlSecSignRsaKey(key)->e);
    }
    if(xmlSecSignRsaKey(key)->d != NULL) {
	rsa->d = BN_dup(xmlSecSignRsaKey(key)->d);
    }

    digestBuf = xmlMalloc(sizeof(unsigned char) * RSA_size(rsa));
    if(digestBuf == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to allocate digest for the key\n",
	    func);	
#endif 	    
	RSA_free(rsa);
	return(-1);
    }
    
    if(digest->digest != NULL) {
	memset(digest->digest, 0, digest->digestSize);
	xmlFree(digest->digest);  
    }    
    digest->digest = digestBuf;
    digest->digestSize = RSA_size(rsa);
        
    if(xmlSecSignRsaSha1ContextRsa(transform) != NULL) {
	RSA_free(xmlSecSignRsaSha1ContextRsa(transform));
    }    
    transform->binData = rsa;
    return(0);
}

/**
 * RSA key
 */
/**
 * xmlSecSignRsaKeyCreate
 * @id:
 *
 */
static xmlSecKeyPtr	
xmlSecSignRsaKeyCreate(xmlSecKeyId id) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecSignRsaKeyCreate";
    xmlSecKeyPtr key;
    
    if(id != xmlSecSignRsaKey) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: id is unknown\n",
	    func);
#endif 	    
	return(NULL);	
    }
    
    key = (xmlSecKeyPtr)xmlMalloc(sizeof(struct _xmlSecKey));
    if(key == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: memory allocation failed\n",
	    func);	
#endif 	    
	return(NULL);
    }
    memset(key, 0, sizeof(struct _xmlSecKey));  
        
    key->id = id;
    return(key);
}

/**
 * xmlSecSignRsaKeyDestroy
 * @key
 *
 */
static void
xmlSecSignRsaKeyDestroy(xmlSecKeyPtr key) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecSignRsaKeyDestroy";

    if(!xmlSecKeyCheckId(key, xmlSecSignRsaKey)) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: key is invalid\n",
	    func);	
#endif 	    
	return;
    }
    
    if(xmlSecSignRsaKey(key) != NULL) {
	RSA_free(xmlSecSignRsaKey(key));
    }    
    memset(key, 0, sizeof(struct _xmlSecKey));
    
    xmlFree(key);		    
}

int		
xmlSecSignRsaKeySet(xmlSecKeyPtr key, RSA *rsa) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecSignRsaKeySet";
    
    if(!xmlSecKeyCheckId(key, xmlSecSignRsaKey)) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: key is invalid or context\n",
	    func);	
#endif 	    
	return(-1);
    }

    if(xmlSecSignRsaKey(key) != NULL) {
	RSA_free(xmlSecSignRsaKey(key));	
    }    
    
    key->keyData = rsa;
    if(rsa->d != NULL) {
	key->type = xmlSecKeyTypePrivate;
    } else {
	key->type = xmlSecKeyTypePublic;
    }
    
    return(0);
}

/**
 * xmlSecSignRsaKeyGenerate
 * @key:
 * @context:
 *
 */
static int		
xmlSecSignRsaKeyGenerate(xmlSecKeyPtr key, void *context) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecSignRsaKeyGenerate";
    RSA *rsa;
    
    if(!xmlSecKeyCheckId(key, xmlSecSignRsaKey)) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: key is invalid or context\n",
	    func);	
#endif 	    
	return(-1);
    }
    
    rsa = RSA_generate_key(1024, 3, NULL, NULL); 
    if(rsa == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: RSA_generate_parameters failed\n",
	    func);	
#endif 	    
	return(-1);    
    }

    if(xmlSecSignRsaKey(key) != NULL) {
	RSA_free(xmlSecSignRsaKey(key));
    }    
    key->keyData = rsa;
    key->type = xmlSecKeyTypePrivate;    
    return(0);    
}

/**
 * xmlSecSignRsaKeyRead
 * @key:
 * @node:
 *
 * http://www.w3.org/TR/xmldsig-core/#sec-RSAKeyValue
 * The RSAKeyValue Element
 *
 * RSA key values have two fields: Modulus and Exponent.
 *
 * <RSAKeyValue>
 *   <Modulus>xA7SEU+e0yQH5rm9kbCDN9o3aPIo7HbP7tX6WOocLZAtNfyxSZDU16ksL6W
 *     jubafOqNEpcwR3RdFsT7bCqnXPBe5ELh5u4VEy19MzxkXRgrMvavzyBpVRgBUwUlV
 *   	  5foK5hhmbktQhyNdy/6LpQRhDUDsTvK+g9Ucj47es9AQJ3U=
 *   </Modulus>
 *   <Exponent>AQAB</Exponent>
 * </RSAKeyValue>
 *
 * Arbitrary-length integers (e.g. "bignums" such as RSA moduli) are 
 * represented in XML as octet strings as defined by the ds:CryptoBinary type.
 *
 * Schema Definition:
 * 
 * <element name="RSAKeyValue" type="ds:RSAKeyValueType"/>
 * <complexType name="RSAKeyValueType">
 *   <sequence>
 *     <element name="Modulus" type="ds:CryptoBinary"/> 
 *     <element name="Exponent" type="ds:CryptoBinary"/>
 *   </sequence>
 * </complexType>
 *
 * DTD Definition:
 * 
 * <!ELEMENT RSAKeyValue (Modulus, Exponent) > 
 * <!ELEMENT Modulus (#PCDATA) >
 * <!ELEMENT Exponent (#PCDATA) >
 *
 * ============================================================================
 * 
 * To support reading/writing private keys an PrivateExponent element is added
 * to the end
 */
static int
xmlSecSignRsaKeyRead(xmlSecKeyPtr key, xmlNodePtr node) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecSignRsaKeyRead";
    xmlNodePtr cur;
    RSA *rsa;
    int privateKey = 0;
    
    if(!xmlSecKeyCheckId(key, xmlSecSignRsaKey) || (node == NULL)) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: key is invalid or node is null\n",
	    func);	
#endif 	    
	return(-1);
    }

    rsa = RSA_new();
    if(rsa == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create rsa key\n",
	    func);	
#endif 	    
	return(-1);
    }    
    cur = xmlSecGetNextElementNode(node->children);
    
    /* first is Modulus node. It is REQUIRED because we do not support Seed and PgenCounter*/
    if((cur == NULL) || (!xmlSecCheckNodeName(cur,  BAD_CAST "Modulus", xmlSecDSigNs))) {
#ifdef DEBUG_XMLSEC
	xmlGenericError(xmlGenericErrorContext,
	    "%s: required element \"Modulus\" missed\n",
	    func);
#endif	    
	RSA_free(rsa);	
	return(-1);
    }
    if(xmlSecNodeGetBNValue(cur, &(rsa->n)) == NULL) {
#ifdef DEBUG_XMLSEC    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to convert element \"Modulus\" value\n",
	    func);
#endif	    
	RSA_free(rsa);
	return(-1);
    }
    cur = xmlSecGetNextElementNode(cur->next);

    /* next is Exponent node. It is REQUIRED because we do not support Seed and PgenCounter*/
    if((cur == NULL) || (!xmlSecCheckNodeName(cur, BAD_CAST "Exponent", xmlSecDSigNs))) {
#ifdef DEBUG_XMLSEC
	xmlGenericError(xmlGenericErrorContext,
	    "%s: required element \"Exponent\" missed\n",
	    func);
#endif	    
	RSA_free(rsa);
	return(-1);
    }
    if(xmlSecNodeGetBNValue(cur, &(rsa->e)) == NULL) {
#ifdef DEBUG_XMLSEC    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to convert element \"Exponent\" value\n",
	    func);
#endif	    
	RSA_free(rsa);
	return(-1);
    }
    cur = xmlSecGetNextElementNode(cur->next);

    if((cur != NULL) && (xmlSecCheckNodeName(cur, BAD_CAST "PrivateExponent", xmlSecDSigNs))) {
        /* next is X node. It is REQUIRED for private key but
	 * we are not sure exactly what do we read */
	if(xmlSecNodeGetBNValue(cur, &(rsa->d)) == NULL) {
#ifdef DEBUG_XMLSEC    
	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to convert element \"PrivateExponent\" value\n",
		func);
#endif	    
	    RSA_free(rsa);
	    return(-1);
	}
	privateKey = 1;
	cur = xmlSecGetNextElementNode(cur->next);
    }

    if(cur != NULL) {
#ifdef DEBUG_XMLSEC    
	 xmlGenericError(xmlGenericErrorContext,
		"%s: unexpected node found\n");
#endif		
	RSA_free(rsa);
	return(-1);
    }

    if(xmlSecSignRsaKey(key) != NULL) {
	RSA_free(xmlSecSignRsaKey(key));
    }    
    key->keyData = rsa;
    if(privateKey) {
	key->type = xmlSecKeyTypePrivate;    
    } else {
	key->type = xmlSecKeyTypePublic;    	
    }
    return(0);
}

/**
 * xmlSecSignRsaKeyWrite
 * @key
 * @type
 * @parent
 *
 */
static int
xmlSecSignRsaKeyWrite(xmlSecKeyPtr key, xmlSecKeyType type, xmlNodePtr parent) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecSignRsaKeyWrite";
    xmlNodePtr cur;
    int ret;
    
    if(!xmlSecKeyCheckId(key, xmlSecSignRsaKey) || (parent == NULL)) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: key is invalid or parent is null\n",
	    func);	
#endif 	    
	return(-1);
    }
    

    /* first is Modulus node */
    cur = xmlSecAddChild(parent, BAD_CAST "Modulus", xmlSecDSigNs);
    if(cur == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
    	    "%s: failed to create \"Modulus\" node\n",
	    func);
#endif 	    
	return(-1);	
    }
    ret = xmlSecNodeSetBNValue(cur, xmlSecSignRsaKey(key)->n, 1);
    if(ret < 0) {
#ifdef DEBUG_XMLSEC    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to convert element \"Modulus\" value\n",
	    key);
#endif	    
	return(-1);
    }    

    /* next is Exponent node. */
    cur = xmlSecAddChild(parent, BAD_CAST "Exponent", xmlSecDSigNs);
    if(cur == NULL) {
#ifdef DEBUG_XMLSEC 
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create \"Exponent\" node\n",
	    func);
#endif 	    
	return(-1);	
    }
    ret = xmlSecNodeSetBNValue(cur, xmlSecSignRsaKey(key)->e, 1);
    if(ret < 0) {
#ifdef DEBUG_XMLSEC    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to convert element \"Exponent\" value\n",
	    func);
#endif	    
	return(-1);
    }

    /* next is PrivateExponent node: write it ONLY for private keys and ONLY if it is requested */
    if(((type == xmlSecKeyTypePrivate) || (type == xmlSecKeyTypeAny)) && 
        (key->type == xmlSecKeyTypePrivate)) { 
	cur = xmlSecAddChild(parent, BAD_CAST "PrivateExponent", xmlSecDSigNs);
	if(cur == NULL) {
#ifdef DEBUG_XMLSEC
    	    xmlGenericError(xmlGenericErrorContext,
	        "%s: failed to create \"PrivateExponent\" node\n",
		func);
#endif 	    
	    return(-1);	
	}
	ret = xmlSecNodeSetBNValue(cur, xmlSecSignRsaKey(key)->d, 1);
	if(ret < 0) {
#ifdef DEBUG_XMLSEC    
	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to convert element \"PrivateExponent\" value\n",
		func);
#endif	    
	    return(-1);
	}
    }

    return(0);
}

