/** 
 * XMLSec library
 *
 * X509 support
 *
 * The X509Data  Element (http://www.w3.org/TR/xmldsig-core/#sec-X509Data)
 *
 * An X509Data element within KeyInfo contains one or more identifiers of keys 
 * or X509 certificates (or certificates' identifiers or a revocation list). 
 * The content of X509Data is:
 *
 *  1. At least one element, from the following set of element types; any of these may appear together or more than once iff (if and only if) each instance describes or is related to the same certificate:
 *  2.
 *    * The X509IssuerSerial element, which contains an X.509 issuer 
 *	distinguished name/serial number pair that SHOULD be compliant 
 *	with RFC2253 [LDAP-DN],
 *    * The X509SubjectName element, which contains an X.509 subject 
 *	distinguished name that SHOULD be compliant with RFC2253 [LDAP-DN],
 *    * The X509SKI element, which contains the base64 encoded plain (i.e. 
 *	non-DER-encoded) value of a X509 V.3 SubjectKeyIdentifier extension.
 *    * The X509Certificate element, which contains a base64-encoded [X509v3] 
 *	certificate, and
 *    * Elements from an external namespace which accompanies/complements any 
 *	of the elements above.
 *    * The X509CRL element, which contains a base64-encoded certificate 
 *	revocation list (CRL) [X509v3].
 *
 * Any X509IssuerSerial, X509SKI, and X509SubjectName elements that appear 
 * MUST refer to the certificate or certificates containing the validation key.
 * All such elements that refer to a particular individual certificate MUST be 
 * grouped inside a single X509Data element and if the certificate to which 
 * they refer appears, it MUST also be in that X509Data element.
 *
 * Any X509IssuerSerial, X509SKI, and X509SubjectName elements that relate to 
 * the same key but different certificates MUST be grouped within a single 
 * KeyInfo but MAY occur in multiple X509Data elements.
 *
 * All certificates appearing in an X509Data element MUST relate to the 
 * validation key by either containing it or being part of a certification 
 * chain that terminates in a certificate containing the validation key.
 *
 * No ordering is implied by the above constraints.
 *
 * Note, there is no direct provision for a PKCS#7 encoded "bag" of 
 * certificates or CRLs. However, a set of certificates and CRLs can occur 
 * within an X509Data element and multiple X509Data elements can occur in a 
 * KeyInfo. Whenever multiple certificates occur in an X509Data element, at 
 * least one such certificate must contain the public key which verifies the 
 * signature.
 *
 * Schema Definition
 *
 *  <element name="X509Data" type="ds:X509DataType"/> 
 *  <complexType name="X509DataType">
 *    <sequence maxOccurs="unbounded">
 *      <choice>
 *        <element name="X509IssuerSerial" type="ds:X509IssuerSerialType"/>
 *        <element name="X509SKI" type="base64Binary"/>
 *        <element name="X509SubjectName" type="string"/>
 *        <element name="X509Certificate" type="base64Binary"/>
 *        <element name="X509CRL" type="base64Binary"/>
 *        <any namespace="##other" processContents="lax"/>
 *      </choice>
 *    </sequence>
 *  </complexType>
 *  <complexType name="X509IssuerSerialType"> 
 *    <sequence> 
 *       <element name="X509IssuerName" type="string"/> 
 *       <element name="X509SerialNumber" type="integer"/> 
 *     </sequence>
 *  </complexType>
 *
 *  DTD
 *
 *    <!ELEMENT X509Data ((X509IssuerSerial | X509SKI | X509SubjectName |
 *                          X509Certificate | X509CRL)+ %X509.ANY;)>
 *    <!ELEMENT X509IssuerSerial (X509IssuerName, X509SerialNumber) >
 *    <!ELEMENT X509IssuerName (#PCDATA) >
 *    <!ELEMENT X509SubjectName (#PCDATA) >
 *    <!ELEMENT X509SerialNumber (#PCDATA) >
 *    <!ELEMENT X509SKI (#PCDATA) >
 *    <!ELEMENT X509Certificate (#PCDATA) >
 *    <!ELEMENT X509CRL (#PCDATA) >
 *
 * See Copyright for the status of this software.
 * 
 * Author: Aleksey Sanin <aleksey@aleksey.com>
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <libxml/tree.h>
#include <openssl/evp.h>
#include <openssl/x509.h>
#include <openssl/x509v3.h>

#include <xmlsec/xmlsec.h>
#include <xmlsec/xmltree.h>
#include <xmlsec/keys.h>
#include <xmlsec/base64.h>
#include <xmlsec/dsa.h>
#include <xmlsec/rsa.h>
#include <xmlsec/x509.h>


#define xmlSecX509Verify(ctx, cert, chain) \
    (((ctx)->x509VerifyCallback != NULL ) ? \
	((ctx)->x509VerifyCallback)((ctx)->x509Context, (cert), (chain)) : \
	-1)
#define xmlSecX509Find(ctx, subj, issuerName, issuerSerial, ski) \
    (((ctx)->x509FindCallback != NULL ) ? \
	((ctx)->x509FindCallback)((ctx)->x509Context, (subj), \
			          (issuerName), (issuerSerial), \
				  (ski)) : \
	NULL)
#define xmlSecX509AddCrl(ctx, crl) \
    (((ctx)->x509AddCRLCallback != NULL ) ? \
	((ctx)->x509AddCRLCallback)((ctx)->x509Context, (crl)) : \
	-1)


static X509*		xmlSecX509IssuerSerialNodeRead		(xmlNodePtr X509IssuerSerialNode, 
			    					 const xmlSecKeysReadContext *ctx);
static X509*		xmlSecX509CertificateRead		(const xmlNodePtr X509Certificate);
static int		xmlSecX509CertificateWrite		(xmlNodePtr X509Certificate,
								 X509 *cert);
static X509_CRL*	xmlSecX509CRLNodeRead			(const xmlNodePtr X509CRLNode);


static xmlSecKeyPtr	xmlSecX509GetCertKey			(X509 *cert, 
								 xmlSecTransformId id);
static int 		xmlSecX509FindNextChainCert		(STACK_OF(X509) *chain, 
								 X509 *cert);


/**
 * xmlSecX509DataNodeRead:
 * @x509DataNode:
 * @id:
 * @type:
 * @ctx:
 *
 *
 */
xmlSecKeyPtr	
xmlSecX509DataNodeRead(const xmlNodePtr x509DataNode, xmlSecTransformId id,
		    xmlSecKeyType type, const xmlSecKeysReadContext *ctx) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecX509DataNodeRead";
    STACK_OF(X509) *chain = NULL;
    X509 *cert = NULL;
    xmlSecKeyPtr key = NULL;
    xmlNodePtr cur;
    int ret;    
    int i;
    
    if((x509DataNode == NULL) || (ctx == NULL)){
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: x509DataNode node or key search context is null\n",
	    func);	
#endif
	return(NULL);
    }
    if(type == xmlSecKeyTypePrivate) {
	/* "by definition" x509 contains *ONLY* public keys */
	return(NULL);
    }
    if((id != xmlSecSignDsaSha1) && (id != xmlSecSignRsaSha1) && 
	(id != xmlSecTransformUnknown)) {
	/* other key types are not supported */
	return(NULL);
    }

    /* init chain */
    chain = sk_X509_new_null();
    if(chain == NULL) {
#ifdef DEBUG_XMLSEC
    	xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create x509 certs stack\n",
	    func);	
#endif
	goto done;
    }

    /* step 1: read everything from X509Data node into the x509 certs stack */
    cur = xmlSecGetNextElementNode(x509DataNode->children);
    while(cur != NULL) {
	if(xmlSecCheckNodeName(cur, BAD_CAST "X509Certificate", xmlSecDSigNs)) {
	    cert = xmlSecX509CertificateRead(cur);
	    if(cert == NULL) {
#ifdef DEBUG_XMLSEC
    		xmlGenericError(xmlGenericErrorContext,
		    "%s: failed to read X509Certificate node\n",
		    func);	
#endif
		goto done;
	    }
	    sk_X509_push(chain, cert);
	} else if(xmlSecCheckNodeName(cur, BAD_CAST "X509SubjectName", xmlSecDSigNs)) {
	    xmlChar *content;

	    content = xmlNodeGetContent(cur);
	    if(content == NULL) {
#ifdef DEBUG_XMLSEC
    		xmlGenericError(xmlGenericErrorContext,
		    "%s: failed to read X509SubjectName node\n",
		    func);	
#endif
		goto done;
	    }
	    
	    cert = xmlSecX509Find(ctx, content, NULL, NULL, NULL);
	    xmlFree(content);
	    if(cert != NULL) {
		sk_X509_push(chain, cert); 
	    } else {
		/* it's not an error! */
#ifdef DEBUG_XMLSEC
    		xmlGenericError(xmlGenericErrorContext,
		    "%s: failed to find cert by subject\n",
		    func);	
#endif	
	    }
	} else if(xmlSecCheckNodeName(cur, BAD_CAST "X509IssuerSerial", xmlSecDSigNs)) {
	    cert = xmlSecX509IssuerSerialNodeRead(cur, ctx);
	    if(cert != NULL) {
		sk_X509_push(chain, cert); 
	    } else {
		/* it's not an error! */
#ifdef DEBUG_XMLSEC
    		xmlGenericError(xmlGenericErrorContext,
		    "%s: failed to read X509IssuerSerial node\n",
		    func);	
#endif	
	    }
	} else if(xmlSecCheckNodeName(cur, BAD_CAST "X509SKI", xmlSecDSigNs)) {
	    xmlChar *content;

	    content = xmlNodeGetContent(cur);
	    if(content == NULL) {
#ifdef DEBUG_XMLSEC
    		xmlGenericError(xmlGenericErrorContext,
		    "%s: failed to read X509SKI node\n",
		    func);	
#endif
		goto done;
	    }
	    
	    cert = xmlSecX509Find(ctx, NULL, NULL, NULL, content);
	    xmlFree(content);
	    if(cert != NULL) {
		sk_X509_push(chain, cert); 
	    } else {
		/* it's not an error! */
#ifdef DEBUG_XMLSEC
    		xmlGenericError(xmlGenericErrorContext,
		    "%s: failed to find cert by ski\n",
		    func);	
#endif	
	    }
	} else if(xmlSecCheckNodeName(cur, BAD_CAST "X509CRL", xmlSecDSigNs)) {
	    X509_CRL *crl;
	    
	    crl = xmlSecX509CRLNodeRead(cur);
	    if(crl == NULL) {
#ifdef DEBUG_XMLSEC
    		xmlGenericError(xmlGenericErrorContext,
		    "%s: failed to read X509CRL node\n",
		    func);	
#endif
		goto done;
	    }
	
	    ret = xmlSecX509AddCrl(ctx, crl);
	    if(ret < 0) {
#ifdef DEBUG_XMLSEC
    		xmlGenericError(xmlGenericErrorContext,
		    "%s: failed to add crl\n",
		    func);	
#endif
		X509_CRL_free(crl);
		goto done;
	    }    
	} else {
	    /* laxi schema validation: ignore unknown nodes */
	}
	cur = xmlSecGetNextElementNode(cur->next);
    }

    /* step2: walk thru the stack and find all "end certificates" 
     * (cert subj is not an issuer of another cert) with key for the 
     * transform we need, verify cert
     */
    for(i = 0; i < chain->num; ++i) { 
	cert = ((X509**)(chain->data))[i];
	if(!xmlSecX509FindNextChainCert(chain, cert)) {
	    ret = xmlSecX509Verify(ctx, cert, chain);	    
	    if(ret == 1) {
		key = xmlSecX509GetCertKey(cert, id);
		if(key != NULL) {
	    	    break;
		}
	    } else if(ret < 0) {
#ifdef DEBUG_XMLSEC
    		xmlGenericError(xmlGenericErrorContext,
		    "%s: certificate verification error\n",
		    func);	
#endif
		goto done;
	    }
	}
    }
    if(key == NULL) {
	/* no acceptable cert found */
	goto done;
    }

    /* remember the chain and do not delete it */
    key->x509Chain = chain;
    chain = NULL;

done:
    if(chain != NULL) {
	sk_X509_pop_free(chain, X509_free);
    }
    return(key);
}

/**
 * xmlSecX509IssuerSerialNodeRead:
 * @X509IssuerSerialNode:
 * @ctx:
 *
 */
static X509*
xmlSecX509IssuerSerialNodeRead(xmlNodePtr X509IssuerSerialNode, 
			       const xmlSecKeysReadContext *ctx) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecX509IssuerSerialNodeRead";
    X509 *cert;
    xmlNodePtr cur;
    xmlChar *issuerName;
    xmlChar *issuerSerial;    
    
    if((X509IssuerSerialNode == NULL) || (ctx == NULL)){
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: X509IssuerSerial node  or context is null\n",
	    func);	
#endif
	return(NULL);
    }

    cur = xmlSecGetNextElementNode(X509IssuerSerialNode->children);

    /* the first is required node X509IssuerName */
    if((cur == NULL) || !xmlSecCheckNodeName(cur, BAD_CAST "X509IssuerName", xmlSecDSigNs)) {
#ifdef DEBUG_XMLSEC
	xmlGenericError(xmlGenericErrorContext,
	    "%s: required node \"X509IssuerName\" missed\n",
	    func);
#endif	    
	return(NULL);
    }    
    issuerName = xmlNodeGetContent(cur);
    if(issuerName == NULL) {
#ifdef DEBUG_XMLSEC
	xmlGenericError(xmlGenericErrorContext,
	    "%s: the \"X509IssuerName\" node content is null\n",
	    func);
#endif	    
	return(NULL);
    }
    cur = xmlSecGetNextElementNode(cur->next); 

    /* next is required node X509SerialNumber */
    if((cur == NULL) || !xmlSecCheckNodeName(cur, BAD_CAST "X509SerialNumber", xmlSecDSigNs)) {
#ifdef DEBUG_XMLSEC
	xmlGenericError(xmlGenericErrorContext,
	    "%s: required node \"X509SerialNumber\" missed\n",
	    func);
#endif	    
	xmlFree(issuerName);
	return(NULL);
    }    
    issuerSerial = xmlNodeGetContent(cur);
    if(issuerSerial == NULL) {
#ifdef DEBUG_XMLSEC
	xmlGenericError(xmlGenericErrorContext,
	    "%s: the \"X509SerialNumber\" node content is null\n",
	    func);
#endif	    
	xmlFree(issuerName);
	return(NULL);
    }
    cur = xmlSecGetNextElementNode(cur->next); 

    if(cur != NULL) {
#ifdef DEBUG_XMLSEC    
	 xmlGenericError(xmlGenericErrorContext,
		"%s: unexpected node found\n");
#endif		
	xmlFree(issuerSerial);
	xmlFree(issuerName);
	return(NULL);
    }
    

    cert = xmlSecX509Find(ctx, NULL, issuerName, issuerSerial, NULL);
    xmlFree(issuerSerial);
    xmlFree(issuerName);
    if(cert == NULL) {
#ifdef DEBUG_XMLSEC
    	xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to find cert\n",
	    func);	
#endif	
	return(NULL);
    }    
    
    return(cert);
}

/**
 * xmlSecX509DataNodeWrite:
 * @x509DataNode:
 * @key:
 *
 */
int
xmlSecX509DataNodeWrite(xmlNodePtr x509DataNode, const xmlSecKeyPtr key) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecX509DataNodeWrite";
    X509 *cert;
    xmlNodePtr cur;
    int ret;
    int i;
    
    if((x509DataNode == NULL) || (key == NULL)){
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: X509Data node or key is null\n",
	    func);	
#endif
	return(-1);
    }
    
    /* remove all existing certs */ 
    xmlNodeSetContent(x509DataNode, NULL);
    if(key->x509Chain == NULL) {
	/* nothing to write */
	return(0);
    }
    
    for(i = 0; i < key->x509Chain->num; ++i) {
	cert = ((X509**)(key->x509Chain->data))[i];
	
	/* todo: write other stuff */
	cur = xmlSecAddChild(x509DataNode, BAD_CAST "X509Certificate", xmlSecDSigNs);
	if(cur == NULL) {
#ifdef DEBUG_XMLSEC
    	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to create X509Certificate node\n",
		func);	
#endif
	    return(-1);	
	}
	ret = xmlSecX509CertificateWrite(cur, cert);
	if(ret < 0) {
#ifdef DEBUG_XMLSEC
    	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to write X509Certificate node\n",
		func);	
#endif
	    return(-1);	
	}	
    }
    return(0);
}

/** 
 * xmlSecX509CertificateRead:
 * @cur:
 *
 */
static X509*		
xmlSecX509CertificateRead(const xmlNodePtr X509CertificateNode) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecX509CertificateRead";
    X509 *cert = NULL;
    BIO *mem = NULL;
    xmlChar *content;
    int ret;

    if(X509CertificateNode == NULL){
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: X509Certificate node is null\n",
	    func);	
#endif
	return(NULL);
    }

    content = xmlNodeGetContent(X509CertificateNode);
    if(content == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to get X509Certificate node content\n",
	    func);	
#endif
	goto done;
    }
    
    mem = BIO_new(BIO_s_mem());
    if(mem == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create mem BIO\n",
	    func);	
#endif
	goto done;
    }
    
    /* usual trick with base64 decoding "in-place" */
    ret = xmlSecBase64Decode(content, (unsigned char*)content, 
			     xmlStrlen(content)); 
    if(ret < 0) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: base64 failed\n",
	    func);	
#endif	
	goto done;
    }
    
    ret = BIO_write(mem, content, ret);
    if(ret <= 0) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: mem BIO write failed\n",
	    func);	
#endif	
	goto done;
    }
    
    
    cert = d2i_X509_bio(mem, NULL);
    if(cert == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to read cert from mem BIO\n",
	    func);	
#endif	
	goto done;
    }
    
done:
    if(content != NULL) {
	xmlFree(content);    
    }
    if(mem != NULL) {
	BIO_free_all(mem);
    }
    return(cert);
}

/**
 * xmlSecX509CertificateWrite
 * @X509CertificateNode:
 * @cert:
 *
 */
static int
xmlSecX509CertificateWrite(xmlNodePtr X509CertificateNode, X509 *cert) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecX509CertificateWrite";
    BIO *mem = NULL;
    BIO *b64 = NULL;
    char *p;
    long size;
    
    if((X509CertificateNode == NULL) || (cert == NULL)){
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: X509Certificate node or cert is null\n",
	    func);	
#endif
	return(-1);
    }

    mem = BIO_new(BIO_s_mem());
    if(mem == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create mem BIO\n",
	    func);	
#endif
	goto done;
    }

    b64 = BIO_new(BIO_f_base64());
    if(b64 == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create b64 BIO\n",
	    func);	
#endif
	goto done;
    }
    
    b64 = BIO_push(b64, mem);
    /* todo: */
    i2d_X509_bio(b64, cert);

    BIO_flush(b64);
        
    p = NULL;
    size = BIO_get_mem_data(mem, &p);
    if((size <= 0) || (p == NULL)){
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: unable to get buffer from bio\n",
	    func);	
#endif
	goto done;
    }
    
    xmlNodeSetContent(X509CertificateNode, BAD_CAST "\n");
    xmlNodeAddContentLen(X509CertificateNode, BAD_CAST p, size);
    
done:
    if(b64 != NULL) {
	BIO_free_all(b64);
    } else if(mem != NULL) {
	BIO_free_all(mem);
    }
    return(0);
}

/**
 * xmlSecX509CRLNodeRead:
 *
 *
 *
 */
static X509_CRL*	
xmlSecX509CRLNodeRead(const xmlNodePtr X509CRLNode) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecX509CRLNodeRead";
    X509_CRL *crl = NULL;
    BIO *mem = NULL;
    xmlChar *content;
    int ret;

    if(X509CRLNode == NULL){
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: X509CRL node is null\n",
	    func);	
#endif
	return(NULL);
    }

    content = xmlNodeGetContent(X509CRLNode);
    if(content == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to get X509CRL node content\n",
	    func);	
#endif
	goto done;
    }
    
    mem = BIO_new(BIO_s_mem());
    if(mem == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create mem BIO\n",
	    func);	
#endif
	goto done;
    }
    
    /* usual trick with base64 decoding "in-place" */
    ret = xmlSecBase64Decode(content, (unsigned char*)content, 
			     xmlStrlen(content)); 
    if(ret < 0) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: base64 failed\n",
	    func);	
#endif	
	goto done;
    }
    
    ret = BIO_write(mem, content, ret);
    if(ret <= 0) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: mem BIO write failed\n",
	    func);	
#endif	
	goto done;
    }
    
    
    crl = d2i_X509_CRL_bio(mem, NULL);
    if(crl == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to read crl from mem BIO\n",
	    func);	
#endif	
	goto done;
    }
    
done:
    if(content != NULL) {
	xmlFree(content);    
    }
    if(mem != NULL) {
	BIO_free_all(mem);
    }
    return(crl);
}


/** 
 * xmlSecX509GetCertKey
 * @cert:
 * @id:
 *
 *
 */
static xmlSecKeyPtr	
xmlSecX509GetCertKey(X509 *cert, xmlSecTransformId id) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecX509GetCertKey";
    xmlSecKeyPtr key = NULL;
    EVP_PKEY *pKey = NULL;
    
    if(cert == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: cert is null\n",
	    func);	
#endif
	return(NULL);
    }

    pKey = X509_get_pubkey(cert);
    if(pKey == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to get public key from cert\n",
	    func);	
#endif
	return(NULL);
    }    

    key = xmlSecKeyReadEvpKey(pKey, id, xmlSecKeyOriginX509);
    if(key != NULL) {
	/* remember cert */    
	key->x509Cert = cert;
    }

    EVP_PKEY_free(pKey);
    return(key);
}

/** 
 * xmlSecX509FindNextChainCert:
 * @chain:
 * @cert:
 *
 *
 */
static int
xmlSecX509FindNextChainCert(STACK_OF(X509) *chain, X509 *cert) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecX509FindNextChainCert";
    unsigned long certSubjHash;
    int i;
    
    if((chain == NULL) || (cert == NULL)) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: chain or cert is null\n",
	    func);	
#endif
	return(-1);
    }
    
    certSubjHash = X509_subject_name_hash(cert);
    for(i = 0; i < chain->num; ++i) {
	if(X509_issuer_name_hash(((X509**)(chain->data))[i]) == certSubjHash) {
	    return(i);
	}
    }
    return(0);
}

/**
 * xmlSecDebugx509Dump
 *
 */
void
xmlSecDebugx509Dump(FILE *output, X509 *cert, STACK_OF(X509) *chain) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecDebugx509Dump";
    char buf[1024];
    BIGNUM *bn = NULL;
    
    if((output == NULL) || (cert == NULL)) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: cert or output file is null\n", 
	    func);	
#endif
	return;
    }
    
    fprintf(output, "=== x509 Certificate\n");
    fprintf(output, "==== Subject Name: %s\n", 
	 X509_NAME_oneline(X509_get_subject_name(cert), buf, sizeof(buf))); 
    fprintf(output, "==== Issuer Name: %s\n", 
	 X509_NAME_oneline(X509_get_issuer_name(cert), buf, sizeof(buf))); 
    fprintf(output, "==== Serial: ");
    bn = ASN1_INTEGER_to_BN(X509_get_serialNumber(cert),NULL);
    if(bn != NULL) {
	BN_print_fp(output, bn);
	BN_free(bn);
	fprintf(output, "\n");
    } else {
	fprintf(output, "unknown\n");
    }
    if(chain != NULL) {
	int i;
	for(i = 0; i < chain->num; ++i) {
	    if(((X509**)(chain->data))[i] != cert) {
		xmlSecDebugx509Dump(output, ((X509**)(chain->data))[i], NULL);
	    }
	}
    }
}

/**
 * xmlSecDerX509Read
 *
 *
 */
xmlSecKeyPtr	
xmlSecDerX509Read(unsigned char *buf, size_t size,
		  xmlSecTransformId id, xmlSecKeyType type,
		  const xmlSecKeysReadContext *ctx) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecDerX509Read";
    X509 *cert = NULL;
    BIO *mem = NULL;
    xmlSecKeyPtr key = NULL;
    int ret;

    if((buf == NULL) || (ctx == NULL)){
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: buffer or context is null\n",
	    func);	
#endif
	return(NULL);
    }
    if(type == xmlSecKeyTypePrivate) {
	/* "by definition" x509 contains *ONLY* public keys */
	return(NULL);
    }

    mem = BIO_new(BIO_s_mem());
    if(mem == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create mem BIO\n",
	    func);	
#endif
	goto done;
    }
    
    ret = BIO_write(mem, buf, size);
    if(ret <= 0) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: mem BIO write failed\n",
	    func);	
#endif	
	goto done;
    }
    
    cert = d2i_X509_bio(mem, NULL);
    if(cert == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to read cert from mem BIO\n",
	    func);	
#endif	
	goto done;
    }
    
    /* verify cert */    
    ret = xmlSecX509Verify(ctx, cert, NULL);	    
    if(ret == 1) {
	key = xmlSecX509GetCertKey(cert, id);
	if(key == NULL) {
	    X509_free(cert);
	}
    } else if(ret < 0) {
#ifdef DEBUG_XMLSEC
    	xmlGenericError(xmlGenericErrorContext,
	    "%s: certificate verification error\n",
	    func);	
#endif
	X509_free(cert);
	goto done;
    } else {
	 X509_free(cert);   
    }
    
    
done:
    if(mem != NULL) {
	BIO_free_all(mem);
    }
    return(key);    
}

