/** 
 * XMLSec library
 *
 * "XML Digital Signature" implementation
 *  http://www.w3.org/TR/xmldsig-core/
 *  http://www.w3.org/Signature/Overview.html
 * 
 * See Copyright for the status of this software.
 * 
 * Author: Aleksey Sanin <aleksey@aleksey.com>
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <libxml/tree.h>
#include <libxml/parser.h> 

#include <xmlsec/xmlsec.h>
#include <xmlsec/xmltree.h>
#include <xmlsec/keys.h>
#include <xmlsec/transforms.h>
#include <xmlsec/digests.h>
#include <xmlsec/membuf.h>

#include <xmlsec/xmldsig.h>

#define xmlDSigSignatureIsValid(signature) \
    ((( signature ) != NULL) && ((signature)->ctx != NULL))
#define xmlDSigGetKeysContext(signature) \
    ((( ( signature ) != NULL ) && ( ( signature )->ctx != NULL ) ) ? \
      (( ( signature )->ctx->keysReadContext)) : NULL)

static xmlDSigSignaturePtr	xmlDSigSignatureCreate		(const xmlDSigCtxPtr ctx,
								 xmlNodePtr self,
								 int sign);
static void			xmlDSigSignatureDestroy		(xmlDSigSignaturePtr signature);
static xmlDSigReferencePtr	xmlDSigSignatureAddSignedInfoRef(xmlDSigSignaturePtr signature,
								 xmlDSigReferencePtr ref);
static xmlDSigReferencePtr	xmlDSigSignatureAddManifestRef	(xmlDSigSignaturePtr signature,
								 xmlDSigReferencePtr ref);
static int			xmlDSigSignatureRead		(xmlDSigSignaturePtr signature,
								 xmlNodePtr self,
								 int sign);
static int			xmlDSigSignatureCalculate	(xmlDSigSignaturePtr signature, 
								 int sign,
								 xmlNodePtr signedInfoNode, 
								 xmlSecTransformPtr c14nMethod, 
								 xmlSecTransformPtr signMethod, 
								 xmlNodePtr  signatureValueNode);
static void			xmlDSigDebugSignatureDump	(FILE *output, 
								 xmlDSigSignaturePtr signature);


static int			xmlDSigSignedInfoRead		(xmlDSigSignaturePtr signature,
								 xmlNodePtr self,								 
								 int sign,
								 xmlNodePtr signatureValueNode,
								 xmlNodePtr keyInfoNode);

static xmlDSigReferencePtr	xmlDSigReferenceCreate		(xmlDSigReferenceType type,
								 xmlDSigCtxPtr ctx,
    								 xmlNodePtr self);
static int 			xmlDSigReferenceRead		(xmlDSigReferencePtr ref,
    								 xmlNodePtr self,
								 int sign);
static void			xmlDSigReferenceDestroy		(xmlDSigReferencePtr ref);
static void			xmlDSigReferenceDestroyAll	(xmlDSigReferencePtr ref);
static void			xmlDSigDebugReferenceDump	(FILE *output, 
								 xmlDSigReferencePtr ref);
static void			xmlDSigDebugReferenceDumpAll	(FILE *output, 
								 xmlDSigReferencePtr ref);


static int			xmlDSigObjectRead		(xmlDSigSignaturePtr signature, 
								 xmlNodePtr objectNode,
								 int sign);

static int			xmlDSigManifestRead		(xmlDSigSignaturePtr signature, 
								 xmlNodePtr manifestNode,
								 int sign);

/**
 * DSig generation/validation
 */
/**
 * xmlDSigValidate
 * @ctx:
 * @cur:
 * @signature:
 *
 */
int
xmlDSigValidate(xmlDSigCtxPtr ctx, xmlNodePtr cur, xmlDSigSignaturePtr *signature) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlDSigValidate";
    int ret;
    if((ctx == NULL) || (cur == NULL) || (signature == NULL)) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: context, cur or signature is null\n", 
	    func);	
#endif
	return(-1);	    
    }
    
    if(xmlSecCheckNodeName(cur, BAD_CAST "Signature", xmlSecDSigNs)) {
	xmlDSigSignaturePtr result = xmlDSigSignatureCreate(ctx, cur, 0);
	if(result == NULL) {
#ifdef DEBUG_XMLSEC
    	    xmlGenericError(xmlGenericErrorContext,
	        "%s: unabl to create new result\n", 
	        func);	
#endif			
	    if((*signature) != NULL) {
		xmlDSigSignatureDestroyAll(*signature);
		(*signature) = NULL;
	    } 
	    return(-1);	    		
	}
	
	ret = xmlDSigSignatureRead(result, cur, 0);
	if(ret < 0) {
#ifdef DEBUG_XMLSEC
    	    xmlGenericError(xmlGenericErrorContext,
	        "%s: signature read failed\n", 
	        func);	
#endif			
	    if((*signature) != NULL) {
		xmlDSigSignatureDestroyAll(*signature);
		(*signature) = NULL;
	    } 
	    xmlDSigSignatureDestroyAll(result);
	    return(-1);	    		
	}
	
	/* add new result */
	if((*signature) != NULL) {
	    (*signature)->next = result;
	    result->prev = (*signature);
	}
	(*signature) = result;	
    } else {
        cur = cur->children;
	while(cur != NULL) {
	    ret = xmlDSigValidate(ctx, cur, signature);
	    if(ret < 0) {
#ifdef DEBUG_XMLSEC
    		xmlGenericError(xmlGenericErrorContext,
		    "%s: children validate failed\n", 
		    func);	
#endif			
		if((*signature) != NULL) {
		    xmlDSigSignatureDestroyAll(*signature);
		    (*signature) = NULL;
		} 
		return(-1);	    		
	    }
	    cur = cur->next;
	}
    }
    return(0);
}

/**
 * xmlDSigGenerate
 * @ctx:
 * @cur:
 * @signature:
 *
 */
int
xmlDSigGenerate(xmlDSigCtxPtr ctx, xmlNodePtr cur, xmlDSigSignaturePtr *signature) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlDSigGenerate";
    int ret;
    
    if((ctx == NULL) || (cur == NULL) || (signature == NULL)) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: context, cur or signature is null\n", 
	    func);	
#endif
	return(-1);	    
    }

    if(xmlSecCheckNodeName(cur, BAD_CAST "Signature", xmlSecDSigNs)) {
	xmlDSigSignaturePtr result = xmlDSigSignatureCreate(ctx, cur, 1);
	if(result == NULL) {
#ifdef DEBUG_XMLSEC
    	    xmlGenericError(xmlGenericErrorContext,
	        "%s: unabl to create new result\n", 
	        func);	
#endif			
	    if((*signature) != NULL) {
		xmlDSigSignatureDestroyAll(*signature);
		(*signature) = NULL;
	    } 
	    return(-1);	    		
	}

	ret = xmlDSigSignatureRead(result, cur, 1);
	if(ret < 0) {
#ifdef DEBUG_XMLSEC
    	    xmlGenericError(xmlGenericErrorContext,
	        "%s: signature read  failed\n", 
	        func);	
#endif			
	    if((*signature) != NULL) {
		xmlDSigSignatureDestroyAll(*signature);
		(*signature) = NULL;
	    } 
	    xmlDSigSignatureDestroyAll(result);
	    return(-1);	    		
	}
	
	/* add new result */
	if((*signature) != NULL) {
	    (*signature)->next = result;
	    result->prev = (*signature);
	}
	(*signature) = result;	
    } else {
        cur = cur->children;
	while(cur != NULL) {
	    ret = xmlDSigGenerate(ctx, cur,  signature);
	    if(ret < 0) {
#ifdef DEBUG_XMLSEC
    		xmlGenericError(xmlGenericErrorContext,
		    "%s: children generate failed\n", 
		    func);	
#endif			
		if((*signature) != NULL) {
		    xmlDSigSignatureDestroyAll(*signature);
		    (*signature) = NULL;
		} 
		return(-1);	    		
	    }
	    cur = cur->next;
	}
    }
    return(0);
}

/**
 * DSig result methods
 */
/**
 * xmlDSigSignatureDestroyAll
 * @signature
 * 
 *
 *
 */
void
xmlDSigSignatureDestroyAll(xmlDSigSignaturePtr signature) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlDSigSignatureDestroyAll";
    
    if(signature == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: context or doc is null\n", 
	    func);	
#endif
	return;	    
    }
    
    while(signature->next != NULL) {
	xmlDSigSignatureDestroy(signature->next);
    }    
    while(signature->prev != NULL) {
	xmlDSigSignatureDestroy(signature->prev);
    }    
    xmlDSigSignatureDestroy(signature);
}

/**
 * xmlDSigSignatureCreate
 * @ctx
 * @self:
 * @sign
 *
 */
static xmlDSigSignaturePtr	
xmlDSigSignatureCreate(xmlDSigCtxPtr ctx, xmlNodePtr self, int sign) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlDSigSignatureCreate";
    xmlDSigSignaturePtr signature;
    
    if((ctx == NULL) || (self == NULL)){
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: context or self is null\n", 
	    func);	
#endif
	return(NULL);	    
    }

    /*
     * Allocate a new xmlDSigSignature and fill the fields.
     */
    signature = (xmlDSigSignaturePtr) xmlMalloc(sizeof(xmlDSigSignature));
    if(signature == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: xmlDSigSignature malloc failed\n",
	    func);	
#endif 	    
	return(NULL);
    }
    memset(signature, 0, sizeof(xmlDSigSignature));
    
    signature->ctx = ctx;
    signature->self = self;
    signature->sign = sign;

    return(signature);
}


/**
 * xmlDSigSignatureDestroy:
 * @signature
 *
 *
 */
static void
xmlDSigSignatureDestroy(xmlDSigSignaturePtr signature) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlDSigSignatureDestroy";
    
    if(signature == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: context or doc is null\n", 
	    func);	
#endif
	return;	    
    }

    /* destroy firstSignRef if needed */
    if(signature->firstSignRef != NULL) {
	xmlDSigReferenceDestroyAll(signature->firstSignRef);
    }

    /* destroy firstManifestRef if needed */
    if(signature->firstManifestRef != NULL) {
	xmlDSigReferenceDestroyAll(signature->firstManifestRef);
    }

    /* remove from the chain */
    if(signature->next != NULL) {
	signature->next->prev = signature->prev;
    }
    if(signature->prev != NULL) {
	signature->prev->next = signature->next;
    }
    /* destroy buffe r*/
    if(signature->buffer != NULL) {
	xmlBufferEmpty(signature->buffer);
	xmlBufferFree(signature->buffer);     
    }
    if(signature->signKey != NULL) {
	xmlSecKeyDestroy(signature->signKey, 0);
    }
    memset(signature, 0, sizeof(xmlDSigSignature));
    xmlFree(signature);
}

/**
 * xmlDSigSignatureAddSignedInfoRef:
 * @signature
 * @ref
 *
 *
 */
static xmlDSigReferencePtr
xmlDSigSignatureAddSignedInfoRef(xmlDSigSignaturePtr signature, xmlDSigReferencePtr ref) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlDSigSignatureAddSignedInfoRef";
    
    if(!xmlDSigSignatureIsValid(signature) || (ref == NULL)){
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: result or reference is null\n", 
	    func);	
#endif
	return(NULL);	    
    }
    
    
    /* add to the list */
    ref->prev = signature->lastSignRef;
    if(signature->lastSignRef != NULL) {
        signature->lastSignRef->next = ref;
    }
    signature->lastSignRef = ref;
    if(signature->firstSignRef == NULL) {
	signature->firstSignRef = ref;
    }
    return(ref);
}

/**
 * xmlDSigSignatureAddManifestRef:
 * @signature
 * @ref
 *
 *
 */
static xmlDSigReferencePtr
xmlDSigSignatureAddManifestRef(xmlDSigSignaturePtr signature, xmlDSigReferencePtr ref) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlDSigSignatureAddManifestRef";
    
    if(!xmlDSigSignatureIsValid(signature) || (ref == NULL)){
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: result or reference is null\n", 
	    func);	
#endif
	return(NULL);	    
    }
    
    
    /* add to the list */
    ref->prev = signature->lastManifestRef;
    if(signature->lastManifestRef != NULL) {
        signature->lastManifestRef->next = ref;
    }
    signature->lastManifestRef = ref;
    if(signature->firstManifestRef == NULL) {
	signature->firstManifestRef = ref;
    }
    return(ref);
}

							 
/**
 * DSig context methods
 */
/**
 * xmlDSigCtxCreate
 *
 *
 *
 *
 */
xmlDSigCtxPtr		
xmlDSigCtxCreate(const xmlSecKeysReadContext *keysReadContext) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlDSigCtxCreate";
    xmlDSigCtxPtr ctx;
    
    /*
     * Allocate a new xmlDSigCtx and fill the fields.
     */
    ctx = (xmlDSigCtxPtr) xmlMalloc(sizeof(xmlDSigCtx));
    if(ctx == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: xmlDSigCtx malloc failed\n",
	    func);	
#endif 	    
	return(NULL);
    }
    memset(ctx, 0, sizeof(xmlDSigCtx));
    
    /* by default we process Manifests and store everything */
    ctx->processManifests = 1;
    ctx->storeSignatures = 1;
    ctx->storeReferences = 1;
    ctx->storeManifests = 1;
    ctx->keysReadContext = keysReadContext;
    return(ctx);
}

/**
 * xmlDSigCtxDestroy:
 * @ctx
 *
 *
 *
 */
void
xmlDSigCtxDestroy(xmlDSigCtxPtr ctx) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlDSigCtxDestroy";
    
    if(ctx == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: context is null\n", 
	    func);	
#endif
	return;	    
    }
    memset(ctx, 0, sizeof(xmlDSigCtx));
    xmlFree(ctx);
}



/**
 * xmlDSigSignatureRead:
 * @signature:
 * @sign:
 *
 * The Signature  element (http://www.w3.org/TR/xmldsig-core/#sec-Signature)
 *
 * The Signature element is the root element of an XML Signature. 
 * Implementation MUST generate laxly schema valid [XML-schema] Signature 
 * elements as specified by the following schema:
 *
 * Schema Definition:
 *
 *  <element name="Signature" type="ds:SignatureType"/>
 *  <complexType name="SignatureType">
 *  <sequence> 
 *     <element ref="ds:SignedInfo"/> 
 *     <element ref="ds:SignatureValue"/> 
 *     <element ref="ds:KeyInfo" minOccurs="0"/> 
 *     <element ref="ds:Object" minOccurs="0" maxOccurs="unbounded"/> 
 *     </sequence> <attribute name="Id" type="ID" use="optional"/>
 *  </complexType>
 *    
 * DTD:
 *    
 *  <!ELEMENT Signature (SignedInfo, SignatureValue, KeyInfo?, Object*)  >
 *  <!ATTLIST Signature  
 *      xmlns   CDATA   #FIXED 'http://www.w3.org/2000/09/xmldsig#'
 *      Id      ID  #IMPLIED >
 *
 */
static int			
xmlDSigSignatureRead(xmlDSigSignaturePtr signature, xmlNodePtr self, int sign) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlDSigSignatureRead";
    xmlNodePtr signedInfoNode;
    xmlNodePtr signatureValueNode;
    xmlNodePtr keyInfoNode;
    xmlNodePtr firstObjectNode;
    xmlNodePtr cur;
    int ret;
    
    if(!xmlDSigSignatureIsValid(signature) || (self == NULL)) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: signature or self is null\n", 
	    func);	
#endif
	return(-1);	    
    }
    
    cur = xmlSecGetNextElementNode(self->children);
    
    /* first node is required SignedInfo */
    if((cur == NULL) || (!xmlSecCheckNodeName(cur, BAD_CAST "SignedInfo", xmlSecDSigNs))) {
#ifdef DEBUG_XMLSEC
	xmlGenericError(xmlGenericErrorContext,
	    "%s: required element \"SignedInfo\" missed\n",
	    func);
#endif	    
        return(-1);
    }
    signedInfoNode = cur;
    cur = xmlSecGetNextElementNode(cur->next);

    /* next node is required SignatureValue */
    if((cur == NULL) || (!xmlSecCheckNodeName(cur, BAD_CAST "SignatureValue", xmlSecDSigNs))) {
#ifdef DEBUG_XMLSEC
	xmlGenericError(xmlGenericErrorContext,
	    "%s: required element \"SignatureValue\" missed\n",
	    func);
#endif	    
	return(-1);
    }
    signatureValueNode = cur;
    cur = xmlSecGetNextElementNode(cur->next);

    /* next node is optional KeyInfo */
    if((cur != NULL) && (xmlSecCheckNodeName(cur, BAD_CAST "KeyInfo", xmlSecDSigNs))) {
	keyInfoNode = cur;
	cur = xmlSecGetNextElementNode(cur->next);
    } else{
	keyInfoNode = NULL;
    }

    
    /* next nodes are optional Object */
    firstObjectNode = NULL;
    while((cur != NULL) && (xmlSecCheckNodeName(cur, BAD_CAST "Object", xmlSecDSigNs))) {
	if(firstObjectNode == NULL) {
	    firstObjectNode = cur;
	}

	/* read manifests from objects */
	if(signature->ctx->processManifests) {
	    ret = xmlDSigObjectRead(signature, cur, sign);
	    if(ret < 0) {
#ifdef DEBUG_XMLSEC
		 xmlGenericError(xmlGenericErrorContext,
		    "%s: unexpected node found\n",
		    func);
#endif		
		return(-1);	    	    
	    }
	}
	cur = xmlSecGetNextElementNode(cur->next);
    }
    
    if(cur != NULL) {
#ifdef DEBUG_XMLSEC
	 xmlGenericError(xmlGenericErrorContext,
	    "%s: unexpected node found\n",
	    func);
#endif		
	return(-1);
    }

    /* now we are ready to read SignedInfo node and calculate/verify signature */
    ret = xmlDSigSignedInfoRead(signature, signedInfoNode, sign, 
				signatureValueNode, keyInfoNode);
    if(ret < 0) {
#ifdef DEBUG_XMLSEC
	 xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to read SignedInfo node\n",
	    func);
#endif		
	return(-1);	
    }				
    
    return(0);
}

/**
 * xmlDSigSignatureCalculate:
 * @signature:
 * @signedInfoNode:
 * @c14nMethod:
 * @signMethod:
 * @signatureValueNode:
 *
 *  The way in which the SignedInfo element is presented to the 
 *  canonicalization method is dependent on that method. The following 
 *  applies to algorithms which process XML as nodes or characters:
 *
 *  - XML based canonicalization implementations MUST be provided with 
 *  a [XPath] node-set originally formed from the document containing 
 *  the SignedInfo and currently indicating the SignedInfo, its descendants,
 *  and the attribute and namespace nodes of SignedInfo and its descendant 
 *  elements.
 *
 *  - Text based canonicalization algorithms (such as CRLF and charset 
 *  normalization) should be provided with the UTF-8 octets that represent 
 *  the well-formed SignedInfo element, from the first character to the 
 *  last character of the XML representation, inclusive. This includes 
 *  the entire text of the start and end tags of the SignedInfo element 
 *  as well as all descendant markup and character data (i.e., the text) 
 *  between those tags. Use of text based canonicalization of SignedInfo 
 *  is NOT RECOMMENDED.   	     
 *
 *  =================================
 *  we do not support any non XML based C14N 
 */
static int
xmlDSigSignatureCalculate(xmlDSigSignaturePtr signature, int sign,
		xmlNodePtr signedInfoNode, xmlSecTransformPtr c14nMethod, 
		xmlSecTransformPtr signMethod, xmlNodePtr signatureValueNode) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlDSigSignatureCalculate";
    xmlNodeSetPtr nodeSet = NULL;
    xmlSecTransformStatePtr state = NULL;
    xmlSecTransformPtr memBuffer = NULL;
    int res = -1;
    int ret;
    
    if(!xmlDSigSignatureIsValid(signature) || (signedInfoNode == NULL) || 
      (c14nMethod == NULL) || (signMethod == NULL) || (signatureValueNode == NULL)){
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: signature or something else is null\n", 
	    func);	
#endif
	return(-1);	    
    }
    
    /* this should be done in different way if C14N is binary! */
    nodeSet = xmlSecGetChildNodeSet(signedInfoNode, NULL, 1);
    if(nodeSet == NULL) {
#ifdef DEBUG_XMLSEC
	xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create signedinfo nodes set\n",
	    func);	
#endif
	goto done;
    }

    state = xmlSecTransformStateCreate(signedInfoNode->doc, nodeSet, NULL);
    if(state == NULL){
#ifdef DEBUG_XMLSEC    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create transforms state\n",
	    func);
#endif	    
	goto done;
    }	

    ret = xmlSecTransformStateUpdate(state, c14nMethod);
    if(ret < 0){
#ifdef DEBUG_XMLSEC    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to add c14n method\n",
	    func);
#endif	    
	goto done;
    }

    /* 
     * if requested then insert a memory buffer to capture the digest data 
     */
    if(signature->ctx->storeSignatures) {
	memBuffer = xmlSecTransformCreate(xmlSecMemBuf, 0, 1);
	if(memBuffer == NULL) {
#ifdef DEBUG_XMLSEC    
	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to create mem buffer\n",
		func);
#endif	    
	    goto done;
	}
	ret = xmlSecTransformStateUpdate(state, memBuffer);
	if(ret < 0){
#ifdef DEBUG_XMLSEC    
	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to add mem buffer\n",
		func);
#endif	    
	    goto done;
	}
    }
     
    ret = xmlSecTransformStateUpdate(state, signMethod);
    if(ret < 0){
#ifdef DEBUG_XMLSEC    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to add sign method\n",
	    func);
#endif	    
	goto done;
    }
    
    ret = xmlSecTransformStateFinal(state, xmlSecTransformResultBinary);
    if(ret < 0) {
#ifdef DEBUG_XMLSEC
	xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to finalize transforms\n",
	    func);
#endif
	goto done;
    }
    
    if(sign) {
	ret = xmlSecDigestSignNode(signMethod, signatureValueNode, 1);
	if(ret < 0) {
#ifdef DEBUG_XMLSEC
	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to sign node\n",
		func);
#endif
	    goto done;	
	}
    } else {
	ret = xmlSecDigestVerifyNode(signMethod, signatureValueNode);
	if(ret < 0) {
#ifdef DEBUG_XMLSEC
	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to verify node\n",
		func);
#endif
	    goto done;	
	}
    }
    signature->result = signMethod->status;

    if(memBuffer != NULL) {
	signature->buffer = xmlSecMemBufTransformGetBuffer(memBuffer, 1);
    }
    
    res = 0;
done:    
    if(state != NULL) {
	xmlSecTransformStateDestroy(state);
    }
    if(nodeSet != NULL) {
	xmlXPathFreeNodeSet(nodeSet);
    }
    if(memBuffer != NULL) {
	xmlSecTransformDestroy(memBuffer, 1);
    }
    return(res);
}

/** 
 * xmlDSigDebugSignatureDumpAll:
 * @output:
 * @signature:
 *
 *
 */
void
xmlDSigDebugSignatureDumpAll(FILE *output, xmlDSigSignaturePtr signature) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlDSigDebugSignatureDumpAll";
    xmlDSigSignaturePtr ptr;
    
    if((output == NULL) || (signature == NULL)) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: signature or output file is null\n", 
	    func);	
#endif
	return;
    }
    
    ptr = signature->prev;
    while(ptr != NULL) {
	xmlDSigDebugSignatureDump(output, ptr);
	ptr = ptr->prev;
    }
    xmlDSigDebugSignatureDump(output, signature);
    ptr = signature->next;
    while(ptr != NULL) {
	xmlDSigDebugSignatureDump(output, ptr);
	ptr = ptr->next;
    }
}

/** 
 * xmlDSigDebugSignatureDump:
 * @output:
 * @signature:
 *
 *
 */
static void
xmlDSigDebugSignatureDump(FILE *output, xmlDSigSignaturePtr signature) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlDSigDebugSignatureDump";
    
    if((output == NULL) || (signature == NULL)) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: signature or output file is null\n", 
	    func);	
#endif
	return;
    }
    
    fprintf(output, "= SIGNATURE (%s)\n", 
	    (signature->sign) ? "generate" : "validate");
    fprintf(output, "== result: %s\n", 
	    (signature->result == xmlSecTransformStatusOk) ? "OK" : "FAIL");    
    fprintf(output, "== sign method: %s\n", 
	    (signature->signMethod != NULL) ? 
	    (char*)((signature->signMethod)->href) : "NULL"); 
    if(signature->signKey != NULL) {
	xmlSecDebugKeyDump(output, signature->signKey);
    }
    if(signature->buffer != NULL) {
	fprintf(output, "== start buffer:\n");
	fwrite(xmlBufferContent(signature->buffer), 
	       xmlBufferLength(signature->buffer), 1,
	       output);
	fprintf(output, "\n== end buffer\n");
    }	    
    
    /* print firstSignRef */
    if(signature->firstSignRef != NULL) {
	fprintf(output, "== SIGNED INFO REFERENCES\n");
	xmlDSigDebugReferenceDumpAll(output, signature->firstSignRef);
    }

    /* print firstManifestRef */
    if(signature->firstManifestRef != NULL) {
	fprintf(output, "== MANIFESTS REFERENCES\n");
	xmlDSigDebugReferenceDumpAll(output, signature->firstManifestRef);
    }
}


/** 
 * xmlDSigSignedInfoRead:
 * @signature:
 * @self:
 * @sign:
 * @signatureValueNode:
 * @keyInfoNode:
 *
 * The SignedInfo  Element (http://www.w3.org/TR/xmldsig-core/#sec-SignedInfo)
 * 
 * The structure of SignedInfo includes the canonicalization algorithm, 
 * a signature algorithm, and one or more references. The SignedInfo element 
 * may contain an optional ID attribute that will allow it to be referenced by 
 * other signatures and objects.
 *
 * SignedInfo does not include explicit signature or digest properties (such as
 * calculation time, cryptographic device serial number, etc.). If an 
 * application needs to associate properties with the signature or digest, 
 * it may include such information in a SignatureProperties element within 
 * an Object element.
 *
 * Schema Definition:
 *
 *  <element name="SignedInfo" type="ds:SignedInfoType"/> 
 *  <complexType name="SignedInfoType">
 *    <sequence> 
 *      <element ref="ds:CanonicalizationMethod"/>
 *      <element ref="ds:SignatureMethod"/> 
 *      <element ref="ds:Reference" maxOccurs="unbounded"/> 
 *    </sequence> 
 *    <attribute name="Id" type="ID" use="optional"/> 
 *  </complexType>
 *    
 * DTD:
 *    
 *  <!ELEMENT SignedInfo (CanonicalizationMethod, SignatureMethod,  Reference+) >
 *  <!ATTLIST SignedInfo  Id   ID      #IMPLIED>
 * 
 */
static int
xmlDSigSignedInfoRead(xmlDSigSignaturePtr signature,  xmlNodePtr self,  int sign,
	   	      xmlNodePtr signatureValueNode, xmlNodePtr keyInfoNode) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlDSigSignedInfoRead";
    xmlSecTransformPtr c14nMethod = NULL;
    xmlSecTransformPtr signMethod = NULL;
    xmlNodePtr cur;
    xmlDSigReferencePtr ref;
    int ret;
    int res = -1;
    
    if(!xmlDSigSignatureIsValid(signature) || (self == NULL) || (signatureValueNode == NULL)){
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: signature or self or signatureValueNode is null\n", 
	    func);	
#endif
	return(-1);	    
    }
    cur = xmlSecGetNextElementNode(self->children);
    
    /* first node is required CanonicalizationMethod */
    if((cur == NULL) || (!xmlSecCheckNodeName(cur, BAD_CAST "CanonicalizationMethod", xmlSecDSigNs))) {
#ifdef DEBUG_XMLSEC    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: required element \"CanonicalizationMethod\" missed\n",
	    func);
#endif	    
	goto done;
    }
    c14nMethod = xmlSecTransformNodeRead(cur, xmlSecUsageDSigC14N, 1);
    if(c14nMethod == NULL) {
#ifdef DEBUG_XMLSEC    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to read \"CanonicalizationMethod\"\n",
	    func);
#endif	    
	goto done;
    }
    cur = xmlSecGetNextElementNode(cur->next);

    /* next node is required SignatureMethod */
    if((cur == NULL) || (!xmlSecCheckNodeName(cur, BAD_CAST "SignatureMethod", xmlSecDSigNs))) {
#ifdef DEBUG_XMLSEC    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: required element \"SignatureMethod\" missed\n",
	    func);
#endif	    
	goto done;
    }
    signMethod = xmlSecTransformNodeRead(cur, xmlSecUsageDSigSignature, 1);
    if(signMethod == NULL) {
#ifdef DEBUG_XMLSEC    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to read \"SignatureMethod\"\n",
	    func);
#endif	    
	goto done;
    }
    signature->signMethod = signMethod->id;
    cur = xmlSecGetNextElementNode(cur->next);

    /* now we are ready to read key */
    if((signature->signKey == NULL) && (keyInfoNode != NULL)) {
	signature->signKey = xmlSecKeyInfoRead(keyInfoNode, signMethod->id, 
    		(sign) ? xmlSecKeyTypePrivate : xmlSecKeyTypeAny,
		xmlDSigGetKeysContext(signature));
    }
    
    /* if there is no key in keyInfo then try the findKey callback */
    if((signature->signKey == NULL) && (
			 xmlDSigGetKeysContext(signature)->allowedOrigins & 
			 xmlSecKeyOriginKeyManager)) {
	signature->signKey = xmlSecKeySearcCtxFindKey(
		xmlDSigGetKeysContext(signature),
		signMethod->id, 
		(sign) ? xmlSecKeyTypePrivate : xmlSecKeyTypeAny, 
		NULL);
    }

    if(signature->signKey == NULL) {
#ifdef DEBUG_XMLSEC    
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to find signature key\n",
	    func);
#endif	    
	goto done;
    }
    ret = xmlSecTransformAddKey(signMethod, signature->signKey);
    if(ret < 0) {
#ifdef DEBUG_XMLSEC    
        xmlGenericError(xmlGenericErrorContext,
    	    "%s: failed to add key\n",
	    func);
#endif	    
	goto done;
    }
    if(sign && (keyInfoNode != NULL)) {
	/* update KeyInfo! */
	ret = xmlSecKeyInfoWrite(keyInfoNode, signature->signKey, 
				xmlSecKeyTypePublic);
	if(ret < 0) {
#ifdef DEBUG_XMLSEC    
	    xmlGenericError(xmlGenericErrorContext,
	        "%s: failed to write \"KeyInfo\"\n",
	        func);
#endif	    
	    goto done;
	}	
    }
    
    /* next is Reference nodes (at least one must present!) */
    while((cur != NULL) && xmlSecCheckNodeName(cur, BAD_CAST "Reference", xmlSecDSigNs)) {
	ref = xmlDSigReferenceCreate(xmlDSigSignedInfoReference, 
				     signature->ctx, cur);
	if(ref == NULL) {
#ifdef DEBUG_XMLSEC    
	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to create reference\n",
	        func);
#endif	    
	    goto done;
	}
	
	ret = xmlDSigReferenceRead(ref, cur, sign);
	if(ret < 0) {
#ifdef DEBUG_XMLSEC    
	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to read \"Reference\"\n",
	        func);
#endif	    
	    xmlDSigReferenceDestroy(ref);
	    goto done;
	}
	
	if(xmlDSigSignatureAddSignedInfoRef(signature, ref) == NULL) {
#ifdef DEBUG_XMLSEC    
	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to add reference\n",
	        func);
#endif	    
	    xmlDSigReferenceDestroy(ref);
	    goto done;
	}	


	if((!sign) && (ref->result != xmlSecTransformStatusOk)) {
#ifdef DEBUG_XMLSEC    
	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to validate \"Reference\"\n",
	        func);
#endif	    
	    /* "soft" error */
	    res = 0;
	    goto done;
	}
	cur = xmlSecGetNextElementNode(cur->next); 
    }
    
    if(signature->firstSignRef == NULL) {
#ifdef DEBUG_XMLSEC    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: no firstSignRef found\n",
	    func);
#endif	    
	goto done;
    }

    if(cur != NULL) {
#ifdef DEBUG_XMLSEC    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: found unexpected node\n",
	    func);
#endif	    
	goto done;
    }

    /* calculate signature and write/verify it*/
    ret = xmlDSigSignatureCalculate(signature, sign, self, 
				    c14nMethod, signMethod, 
				    signatureValueNode);
    if(ret < 0) {
#ifdef DEBUG_XMLSEC    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to calculate signature\n",
	    func);
#endif	    
	goto done;
    }				    
    
    /* we are ok! */
    res = 0;    
done:
    if(c14nMethod != NULL) {
	xmlSecTransformDestroy(c14nMethod, 1);
    }
    if(signMethod != NULL) {
	xmlSecTransformDestroy(signMethod, 1);
    }
    return(res);
}


/**
 * xmlDSigReferenceRead:
 * @ref:
 * @self:
 * @sign:
 *
 * The Reference  Element (http://www.w3.org/TR/xmldsig-core/#sec-Reference)
 * 
 * Reference is an element that may occur one or more times. It specifies 
 * a digest algorithm and digest value, and optionally an identifier of the 
 * object being signed, the type of the object, and/or a list of transforms 
 * to be applied prior to digesting. The identification (URI) and transforms 
 * describe how the digested content (i.e., the input to the digest method) 
 * was created. The Type attribute facilitates the processing of referenced 
 * data. For example, while this specification makes no requirements over 
 * external data, an application may wish to signal that the referent is a 
 * Manifest. An optional ID attribute permits a Reference to be referenced 
 * from elsewhere.
 *
 * Schema Definition:
 *
 *  <element name="Reference" type="ds:ReferenceType"/>
 *  <complexType name="ReferenceType">
 *    <sequence> 
 *      <element ref="ds:Transforms" minOccurs="0"/> 
 *      <element ref="ds:DigestMethod"/> 
 *      <element ref="ds:DigestValue"/> 
 *    </sequence>
 *    <attribute name="Id" type="ID" use="optional"/> 
 *    <attribute name="URI" type="anyURI" use="optional"/> 
 *    <attribute name="Type" type="anyURI" use="optional"/> 
 *  </complexType>
 *    
 * DTD:
 *    
 *   <!ELEMENT Reference (Transforms?, DigestMethod, DigestValue)  >
 *   <!ATTLIST Reference Id  ID  #IMPLIED
 *   		URI CDATA   #IMPLIED
 * 		Type    CDATA   #IMPLIED>
 *
 *
 */
static int
xmlDSigReferenceRead(xmlDSigReferencePtr ref, xmlNodePtr self, int sign) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlDSigReferenceRead";
    xmlNodePtr cur;
    xmlSecTransformStatePtr state = NULL;
    xmlSecTransformPtr digestMethod = NULL;
    xmlNodePtr digestValueNode;
    xmlSecTransformPtr memBuffer = NULL;
    int res = -1;
    int ret;
 
    if((ref == NULL) || (self == NULL)){
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: ref or self is null\n", 
	    func);	
#endif
	return(-1);	    
    }
    cur = xmlSecGetNextElementNode(self->children);
    
    /* read attributes first */
    ref->uri = xmlGetProp(self, BAD_CAST "URI");
    ref->id  = xmlGetProp(self, BAD_CAST "Id");
    ref->type= xmlGetProp(self, BAD_CAST "Type");

    state = xmlSecTransformStateCreate(self->doc, NULL, (char*)ref->uri);
    if(state == NULL){
#ifdef DEBUG_XMLSEC    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create transforms state\n",
	    func);
#endif	    
	goto done;
    }	

    /* first is optional Transforms node */
    if((cur != NULL) && xmlSecCheckNodeName(cur, BAD_CAST "Transforms", xmlSecDSigNs)) {
	ret = xmlSecTransformsNodeRead(state, cur);
	if(ret < 0){
#ifdef DEBUG_XMLSEC    
	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to read \"Transforms\"\n",
		func);
#endif	    
	    goto done;
	}	
	cur = xmlSecGetNextElementNode(cur->next);
    }

    /* 
     * if requested then insert a memory buffer to capture the digest data 
     */
    if(ref->ctx->storeReferences) {
	memBuffer = xmlSecTransformCreate(xmlSecMemBuf, 0, 1);
	if(memBuffer == NULL) {
#ifdef DEBUG_XMLSEC    
	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to create mem buffer\n",
		func);
#endif	    
	    goto done;
	}
	ret = xmlSecTransformStateUpdate(state, memBuffer);
	if(ret < 0){
#ifdef DEBUG_XMLSEC    
	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to add mem buffer\n",
		func);
#endif	    
	    goto done;
	}
    }
     
    /* next node is required DigestMethod */
    if((cur == NULL) || (!xmlSecCheckNodeName(cur, BAD_CAST "DigestMethod", xmlSecDSigNs))) {
#ifdef DEBUG_XMLSEC    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: required element \"DigestMethod\" missed\n",
	    func);
#endif	    
	goto done;
    }
    digestMethod = xmlSecTransformNodeRead(cur, xmlSecUsageDSigDigest, 1);
    if(digestMethod == NULL) {
#ifdef DEBUG_XMLSEC    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to read \"digestMethod\"\n",
	    func);
#endif	    
	goto done;
    }
    ret = xmlSecTransformStateUpdate(state, digestMethod);
    if(ret < 0){
#ifdef DEBUG_XMLSEC    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to add \"digestMethod\"\n",
	    func);
#endif	    
	goto done;
    }
    ref->digestMethod = digestMethod->id;
    cur = xmlSecGetNextElementNode(cur->next);

    /* next node is required DigestValue */
    if((cur == NULL) || (!xmlSecCheckNodeName(cur, BAD_CAST "DigestValue", xmlSecDSigNs))) {
#ifdef DEBUG_XMLSEC    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: required element \"DigestValue\" missed\n",
	    func);
#endif	    
	goto done;
    }
    digestValueNode = cur;
    cur = xmlSecGetNextElementNode(cur->next);     

    if(cur != NULL) {
#ifdef DEBUG_XMLSEC    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: found unexpected node\n",
	    func);
#endif	    
	goto done;
    }
    
    ret = xmlSecTransformStateFinal(state, xmlSecTransformResultBinary);
    if(ret < 0) {
#ifdef DEBUG_XMLSEC
	xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to finalize transforms\n",
	    func);
#endif
	goto done;
    }
    
    if(sign) {
	ret = xmlSecDigestSignNode(digestMethod, digestValueNode, 1);
	if(ret < 0) {
#ifdef DEBUG_XMLSEC
	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to sign node\n",
		func);
#endif
	    goto done;	
	}
    } else {
	ret = xmlSecDigestVerifyNode(digestMethod, digestValueNode);
	if(ret < 0) {
#ifdef DEBUG_XMLSEC
	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to verify node\n",
		func);
#endif
	    goto done;	
	}
    }
    ref->result = digestMethod->status;
    
    if(memBuffer != NULL) {
	ref->buffer = xmlSecMemBufTransformGetBuffer(memBuffer, 1);
    }
    res = 0;

done:
    if(state != NULL) {
	xmlSecTransformStateDestroy(state);
    }
    if(digestMethod != NULL) {
	xmlSecTransformDestroy(digestMethod, 1);
    }
    if(memBuffer != NULL) {
	xmlSecTransformDestroy(memBuffer, 1);
    }
    return(res);
}

    


/**
 * xmlDSigReferenceCreate
 * @ctx
 * @self
 *
 *
 */
static xmlDSigReferencePtr	
xmlDSigReferenceCreate(xmlDSigReferenceType type, xmlDSigCtxPtr ctx, xmlNodePtr self) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlDSigReferenceCreate";
    xmlDSigReferencePtr ref;
        
    if((ctx == NULL) || (self == NULL)) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: context or node is null\n", 
	    func);	
#endif
	return(NULL);	    
    }

    /*
     * Allocate a new xmlDSigReference and fill the fields.
     */
    ref = (xmlDSigReferencePtr) xmlMalloc(sizeof(xmlDSigReference));
    if(ref == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: xmlDSigReference malloc failed\n",
	    func);	
#endif 	    
	return(NULL);
    }
    memset(ref, 0, sizeof(xmlDSigReference));
    
    ref->refType = type;
    ref->ctx = ctx;
    ref->self = self;    
    return(ref);
}

/**
 * xmlDSigReferenceDestroy:
 * @ref
 *
 *
 *
 */
static void			
xmlDSigReferenceDestroy(xmlDSigReferencePtr ref) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlDSigReferenceDestroy";
    
    if(ref == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: reference is null\n", 
	    func);	
#endif
	return;	    
    }
    
    /* destroy all strings */
    if(ref->uri) {
	xmlFree(ref->uri);
    }
    if(ref->id) {
	xmlFree(ref->id);
    }
    if(ref->type) {
	xmlFree(ref->type);
    }
    
    /* destroy buffer */
    if(ref->buffer != NULL) {
	xmlBufferEmpty(ref->buffer);
	xmlBufferFree(ref->buffer); 
    }
    
    /* remove from the chain */
    if(ref->next != NULL) {
	ref->next->prev = ref->prev;
    }
    if(ref->prev != NULL) {
	ref->prev->next = ref->next;
    }
    memset(ref, 0, sizeof(xmlDSigReference));
    xmlFree(ref);
}

/**
 * xmlDSigReferenceDestroyAll:
 * @ref:
 *
 *
 */
static void
xmlDSigReferenceDestroyAll(xmlDSigReferencePtr ref) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlDSigReferenceDestroyAll";
    
    if(ref == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: reference is null\n", 
	    func);	
#endif
	return;	    
    }
    while(ref->next != NULL) {
	xmlDSigReferenceDestroy(ref->next);
    }    
    while(ref->prev != NULL) {
	xmlDSigReferenceDestroy(ref->prev);
    }    
    xmlDSigReferenceDestroy(ref);
}

/**
 * xmlDSigDebugReferenceDumpAll:
 * @output:
 * @ref:
 *
 */
static void
xmlDSigDebugReferenceDump(FILE *output, xmlDSigReferencePtr ref) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlDSigDebugReferenceDumpAll";
    
    if((output == NULL) || (ref == NULL)) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: reference or output file is null\n", 
	    func);	
#endif
	return;
    }

    fprintf(output, "=== REFERENCE \n");
    fprintf(output, "==== ref type: %s\n", 
	    (ref->refType == xmlDSigSignedInfoReference) ? 
		"SignedInfo Reference" : "Manifest Reference"); 
    fprintf(output, "==== result: %s\n", 
	    (ref->result == xmlSecTransformStatusOk) ? "OK" : "FAIL");
    fprintf(output, "==== digest method: %s\n", 
	    (ref->digestMethod != NULL) ? (char*)ref->digestMethod->href : "NULL"); 
    fprintf(output, "==== uri: %s\n", 
	    (ref->uri != NULL) ? (char*)ref->uri : "NULL"); 
    fprintf(output, "==== type: %s\n", 
	    (ref->type != NULL) ? (char*)ref->type : "NULL"); 
    fprintf(output, "==== id: %s\n", 
	    (ref->id != NULL) ? (char*)ref->id : "NULL"); 
    
    if(ref->buffer != NULL) {
	fprintf(output, "==== start buffer:\n");
	fwrite(xmlBufferContent(ref->buffer), 
	       xmlBufferLength(ref->buffer), 1,
	       output);
	fprintf(output, "\n==== end buffer:\n");
    }   	    
}

/**
 * xmlDSigDebugReferenceDumpAll:
 * @output:
 * @ref:
 *
 */
static void
xmlDSigDebugReferenceDumpAll(FILE *output, xmlDSigReferencePtr ref) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlDSigDebugReferenceDumpAll";
    xmlDSigReferencePtr ptr;
    
    if((output == NULL) || (ref == NULL)) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: reference or output file is null\n", 
	    func);	
#endif
	return;
    }
    
    ptr = ref->prev;
    while(ptr != NULL) {
	xmlDSigDebugReferenceDump(output, ptr);
	ptr = ptr->prev;
    }
    xmlDSigDebugReferenceDump(output, ref);
    ptr = ref->next;
    while(ptr != NULL) {
	xmlDSigDebugReferenceDump(output, ptr);
	ptr = ptr->next;
    }
}



/**
 * xmlDSigObjectRead:
 * @signature:
 * @objectNode:
 * 
 * The Object Element (http://www.w3.org/TR/xmldsig-core/#sec-Object)
 * 
 * Object is an optional element that may occur one or more times. When 
 * present, this element may contain any data. The Object element may include 
 * optional MIME type, ID, and encoding attributes.
 *     
 * Schema Definition:
 *     
 * <element name="Object" type="ds:ObjectType"/> 
 * <complexType name="ObjectType" mixed="true">
 *   <sequence minOccurs="0" maxOccurs="unbounded">
 *     <any namespace="##any" processContents="lax"/>
 *   </sequence>
 *   <attribute name="Id" type="ID" use="optional"/> 
 *   <attribute name="MimeType" type="string" use="optional"/>
 *   <attribute name="Encoding" type="anyURI" use="optional"/> 
 * </complexType>
 *	
 * DTD:
 *	
 * <!ELEMENT Object (#PCDATA|Signature|SignatureProperties|Manifest %Object.ANY;)* >
 * <!ATTLIST Object  Id  ID  #IMPLIED 
 *                   MimeType    CDATA   #IMPLIED 
 *                   Encoding    CDATA   #IMPLIED >
 */
static int
xmlDSigObjectRead(xmlDSigSignaturePtr signature, xmlNodePtr objectNode, int sign) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlDSigObjectRead";
    xmlNodePtr cur;
    int ret;
    
    if(!xmlDSigSignatureIsValid(signature) || (objectNode == NULL)) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: signature or objectNode is null\n", 
	    func);	
#endif
	return(-1);	    
    }

    cur = xmlSecGetNextElementNode(objectNode->children);
    while(cur != NULL) {
	if(xmlSecCheckNodeName(cur, BAD_CAST "Manifest", xmlSecDSigNs)) {
	    ret = xmlDSigManifestRead(signature, cur, sign);
	    if(ret < 0){
#ifdef DEBUG_XMLSEC
    		xmlGenericError(xmlGenericErrorContext,
		    "%s: Manifest read failed\n", 
		    func);	
#endif
		return(-1);	    
	    }
	}
	cur = xmlSecGetNextElementNode(cur->next);
    }
    return(0);
}

/**
 * xmlDSigManifestRead: 
 * @signature:
 * @manifestNode:
 * @sign:
 *
 * The Manifest  Element (http://www.w3.org/TR/xmldsig-core/#sec-Manifest)
 *
 * The Manifest element provides a list of References. The difference from 
 * the list in SignedInfo is that it is application defined which, if any, of 
 * the digests are actually checked against the objects referenced and what to 
 * do if the object is inaccessible or the digest compare fails. If a Manifest 
 * is pointed to from SignedInfo, the digest over the Manifest itself will be 
 * checked by the core signature validation behavior. The digests within such 
 * a Manifest are checked at the application's discretion. If a Manifest is 
 * referenced from another Manifest, even the overall digest of this two level 
 * deep Manifest might not be checked.
 *     
 * Schema Definition:
 *     
 * <element name="Manifest" type="ds:ManifestType"/> 
 * <complexType name="ManifestType">
 *   <sequence>
 *     <element ref="ds:Reference" maxOccurs="unbounded"/> 
 *   </sequence> 
 *   <attribute name="Id" type="ID" use="optional"/> 
 *  </complexType>
 *	
 * DTD:
 *
 * <!ELEMENT Manifest (Reference+)  >
 * <!ATTLIST Manifest Id ID  #IMPLIED >
 */
static int
xmlDSigManifestRead(xmlDSigSignaturePtr signature, xmlNodePtr manifestNode, int sign) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlDSigManifestRead";
    xmlNodePtr cur;
    xmlDSigReferencePtr ref;
    int ret;
    
    if(!xmlDSigSignatureIsValid(signature) || (manifestNode == NULL)) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: signature or manifestNode is null\n", 
	    func);	
#endif
	return(-1);	    
    }

    cur = xmlSecGetNextElementNode(manifestNode->children);
    while((cur != NULL) && xmlSecCheckNodeName(cur, BAD_CAST "Reference", xmlSecDSigNs)) { 
	ref = xmlDSigReferenceCreate(xmlDSigManifestReference, 
				     signature->ctx, cur);
	if(ref == NULL) {
#ifdef DEBUG_XMLSEC    
	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to create reference\n",
	        func);
#endif	    
	    return(-1);
	}
	
	ret = xmlDSigReferenceRead(ref, cur, sign);
	if(ret < 0) {
#ifdef DEBUG_XMLSEC    
	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to read \"Reference\"\n",
	        func);
#endif	    
	    xmlDSigReferenceDestroy(ref);
	    return(-1);
	}
	
	if(xmlDSigSignatureAddManifestRef(signature, ref) == NULL) {
#ifdef DEBUG_XMLSEC    
	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to add reference\n",
	        func);
#endif	    
	    xmlDSigReferenceDestroy(ref);
	    return(-1);
	}	
	cur = xmlSecGetNextElementNode(cur->next);
    }

    if(cur != NULL) {
#ifdef DEBUG_XMLSEC    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: found unexpected node\n",
	    func);
#endif	    
	return(-1);
    }    
    return(0);
}






/**
 * Creating DSig "on-the-fly"
 */			
/** 
 * xmlDSigSignatureNodeCreate
 * @doc: the parent document
 * @id: the signature Id
 * @c14nMethod: the c14n method to use
 * @c14nData: the c14n method data (see the method header file 
 *            for the format description)
 * @signMethod: the signmethod
 * @signData: the sign method data (see the method header file 
 *            for the format description)
 *
 * The function creates basic sceleton of the signature and returns a pointer
 * to the top level "Signature" node. The caller is responsible
 * for adding this node to the document, calling other xmlDSigXXXXNodeCreate() 
 * functions to finalize the skeleton and calling xmlDSigGenerate()
 * function to actually sign the message.
 *
 * Returns pointer to the "Signature" node or NULL in the case of error.
 *
 */
xmlNodePtr		
xmlDSigSignatureNodeCreate(xmlDocPtr doc, const xmlChar *id,
		       xmlSecTransformId c14nMethod, void *c14nData,
		       xmlSecTransformId signMethod, void *signData) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlDSigSignatureNodeCreate";
    xmlNodePtr signatureNode = NULL;
    xmlNodePtr signedInfoNode;
    xmlNodePtr cur;
    int ret;
    
    if((doc == NULL) || (c14nMethod == xmlSecTransformUnknown) ||
			(signMethod == xmlSecTransformUnknown)) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: doc or one of c14n and sign methods is null\n", 
	    func);	
#endif
	return(NULL);	    
    }
    
    /* creating Signature node */
    signatureNode = xmlNewDocNode(doc, NULL, BAD_CAST "Signature", NULL); 
    if(signatureNode == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s failed to create new Signature node\n",
	    func);
#endif
	return(NULL);	
    }
    if(xmlNewNs(signatureNode, xmlSecDSigNs, NULL) == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to add DSig namespace\n",
	    func);
#endif
	xmlUnlinkNode(signatureNode);
	xmlFreeNode(signatureNode);
	return(NULL);    
    }
    if(id != NULL) {
	xmlSetProp(signatureNode, BAD_CAST "Id", id);
    }

    /* creating SignedInfo node */
    signedInfoNode = xmlSecAddChild(signatureNode, BAD_CAST "SignedInfo", NULL);
    if(signedInfoNode == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s failed to create new SignedInfo node\n",
	    func);
#endif
	xmlUnlinkNode(signatureNode);
	xmlFreeNode(signatureNode);  
	return(NULL);	
    }    
    
    /* creating SignatureValue node */
    cur = xmlSecAddChild(signatureNode, BAD_CAST "SignatureValue", NULL);
    if(cur == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s failed to create new SignatureValue node\n",
	    func);
#endif
	xmlUnlinkNode(signatureNode);
	xmlFreeNode(signatureNode);  
	return(NULL);	
    }
    
    /* now create sub-nodes in SignedInfo node */
    cur = xmlSecAddChild(signedInfoNode, BAD_CAST "CanonicalizationMethod", NULL);
    if(cur == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s failed to create new CanonicalizationMethod node\n",
	    func);
#endif
	xmlUnlinkNode(signatureNode);
	xmlFreeNode(signatureNode);  
	return(NULL);	
    }
    ret = xmlSecTransformNodeWrite(cur, c14nMethod, c14nData);
    if(ret < 0) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s failed to write CanonicalizationMethod node data\n",
	    func);
#endif
	xmlUnlinkNode(signatureNode);
	xmlFreeNode(signatureNode);  
	return(NULL);	
    }

    cur = xmlSecAddChild(signedInfoNode, BAD_CAST "SignatureMethod", NULL);
    if(cur == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s failed to create new SignatureMethod node\n",
	    func);
#endif
	xmlUnlinkNode(signatureNode);
	xmlFreeNode(signatureNode);  
	return(NULL);	
    }
    ret = xmlSecTransformNodeWrite(cur, signMethod, signData);
    if(ret < 0) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s failed to write SignatureMethod node data\n",
	    func);
#endif
	xmlUnlinkNode(signatureNode);
	xmlFreeNode(signatureNode);  
	return(NULL);	
    }
    
    return(signatureNode);
}

/** 
 * xmlDSigReferenceNodeCreate:
 * @signatureNode: the Signature node created by xmlDSigSignatureNodeCreate() 
 *                 function
 * @id: the reference id (may be NULL)
 * @uri: the reference uri 
 * @type: the reference type (may be NULL)
 * @digestMethod: the digest method used in this reference
 * @digestData: the digest method data (see the method header file
 *              for the format description) 
 *
 * Adds new reference to the Signature.
 *
 * Returns pointer to the Reference node or NULL in the case of error
 */
xmlNodePtr
xmlDSigReferenceNodeCreate(xmlNodePtr signatureNode, const xmlChar *id,
		       const xmlChar *uri, const xmlChar *type,
		       xmlSecTransformId digestMethod, void *digestData) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlDSigReferenceNodeCreate";
    xmlNodePtr referenceNode = NULL;
    xmlNodePtr signedInfoNode;
    xmlNodePtr cur;
    int ret;    

    if((signatureNode == NULL) || (digestMethod == xmlSecTransformUnknown)) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: Signature node or digest is null\n", 
	    func);	
#endif
	return(NULL);	    
    }
    
    /* Find SignedInfo node */
    signedInfoNode = xmlSecFindChild(signatureNode, BAD_CAST "SignedInfo", xmlSecDSigNs);
    if(signedInfoNode == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: unable to find SignedInfo node\n", 
	    func);	
#endif
	return(NULL);	    
    }
        
    /* create Reference node */
    referenceNode = xmlSecAddChild(signedInfoNode, BAD_CAST "Reference", NULL);
    if(referenceNode == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s failed to create new Reference node\n",
	    func);
#endif
	return(NULL);	
    }
    if(id != NULL) {
	xmlSetProp(referenceNode, BAD_CAST "Id", id);
    }
    if(uri != NULL) {
	xmlSetProp(referenceNode, BAD_CAST "URI", uri);
    }
    if(type != NULL) {
	xmlSetProp(referenceNode, BAD_CAST "Type", type);
    }

    /* add DigestMethod node */
    cur = xmlSecAddChild(referenceNode, BAD_CAST "DigestMethod", NULL);
    if(cur == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s failed to create new DigestMethod node\n",
	    func);
#endif
	xmlUnlinkNode(referenceNode);
	xmlFreeNode(referenceNode);  
	return(NULL);	
    }
    ret = xmlSecTransformNodeWrite(cur, digestMethod, digestData);
    if(ret < 0) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s failed to write DigestMethod node data\n",
	    func);
#endif
	xmlUnlinkNode(referenceNode);
	xmlFreeNode(referenceNode);  
	return(NULL);	
    }

    /* add DigestValue node */
    cur = xmlSecAddChild(referenceNode, BAD_CAST "DigestValue", NULL);
    if(cur == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s failed to create new DigestValue node\n",
	    func);
#endif
	xmlUnlinkNode(referenceNode);
	xmlFreeNode(referenceNode);  
	return(NULL);	
    }
    
    return(referenceNode);
}

/** 
 * xmlDSigTransformNodeCreate:
 * @referenceNode: the Reference node created by xmlDSigReferenceNodeCreate() function
 * @transformMethod: the new tranform 
 * @transformData: the new transform data (see the method header file  
 *                 for the format description)
 * 
 * Adds new Transform node to the Reference node.
 *
 * Returns pointer to the new Transform node or NULL in the case of error
 */
xmlNodePtr		
xmlDSigTransformNodeCreate(xmlNodePtr referenceNode, 
		       xmlSecTransformId transformMethod, void *transformData) {
    static const char func[] _UNUSED_VARIABLE_ = "";
    xmlNodePtr transformsNode;
    xmlNodePtr transformNode = NULL;
    int ret;
    
    if((referenceNode == NULL) || (transformMethod == xmlSecTransformUnknown)) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: Reference node or tranform is null\n", 
	    func);	
#endif
	return(NULL);	    
    }
    
    transformsNode = xmlSecFindChild(referenceNode, BAD_CAST "Transforms", xmlSecDSigNs);
    if(transformsNode == NULL) {
	xmlNodePtr cur;
	
	cur = xmlSecFindChild(referenceNode, BAD_CAST "DigestMethod", xmlSecDSigNs);
	if(cur == NULL) {
#ifdef DEBUG_XMLSEC
    	    xmlGenericError(xmlGenericErrorContext,
		"%s: unable to find DigestMethod node\n", 
		func);	
#endif
	    return(NULL);	    
	}
	transformsNode = xmlSecAddPrevSibling(cur, BAD_CAST "Transforms", NULL);
	if(transformsNode == NULL) { 
#ifdef DEBUG_XMLSEC
    	    xmlGenericError(xmlGenericErrorContext,
		"%s failed to create new Transforms node\n",
		func);
#endif
	    return(NULL);	
	}
    }

    /* add DigestMethod node */
    transformNode = xmlSecAddChild(transformsNode, BAD_CAST "Transform", NULL);
    if(transformNode == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s failed to create new DigestMethod node\n",
	    func);
#endif
	return(NULL);	
    }
    ret = xmlSecTransformNodeWrite(transformNode, transformMethod, transformData);
    if(ret < 0) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s failed to write Transform node data\n",
	    func);
#endif
	xmlUnlinkNode(transformNode);
	xmlFreeNode(transformNode);  
	return(NULL);	
    }

        
    return(transformNode);
}

/** 
 * xmlDSigKeyInfoNodeCreate:
 * @signatureNode: the Signature node created by xmlDSigSignatureNodeCreate() 
 *                 function
 * @id: the KeyInfo node id (may be NULL)
 * @origin: the set of nodes to add to KeyInfo 
 *
 * Adds new KeyInfo node to Signature node. If there already exist
 * a KeyInfo node then this function adds elements from origin.
 * Attention: to add RetrievalMethod node call special
 * xmlDSigRetrivalMethodCreate() fucntion
 *
 * Returns pointer to the KeyInfo node or NULL in the case of error
 */
xmlNodePtr		
xmlDSigKeyInfoNodeCreate(xmlNodePtr signatureNode, const xmlChar *id,
		     xmlSecKeyOrigin origin) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlDSigKeyInfoNodeCreate";
    xmlNodePtr keyInfoNode = NULL;
    xmlNodePtr cur;
    
    if(signatureNode == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: Signature node is null\n", 
	    func);	
#endif
	return(NULL);	    
    }
    
    keyInfoNode = xmlSecFindChild(signatureNode, BAD_CAST "KeyInfo", xmlSecDSigNs);
    if(keyInfoNode == NULL) {
	xmlNodePtr signatureValueNode;
	
	signatureValueNode = xmlSecFindChild(signatureNode, BAD_CAST "SignatureValue", xmlSecDSigNs);
	if(signatureValueNode == NULL) {
#ifdef DEBUG_XMLSEC
    	    xmlGenericError(xmlGenericErrorContext,
		"%s: unable to find SignatureValue node\n", 
		func);	
#endif
	    return(NULL);	    
	}
        keyInfoNode = xmlSecAddNextSibling(signatureValueNode, BAD_CAST "KeyInfo", NULL);
	if(keyInfoNode == NULL) {
#ifdef DEBUG_XMLSEC
    	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to create KeyInfo node\n", 
		func);	
#endif
	    return(NULL);	    
	}
	if(id == NULL) {
	    xmlSetProp(keyInfoNode, BAD_CAST "Id", id);
	}
    }
    
    /* add sub elements */
    if(origin & xmlSecKeyOriginKeyName) {
	cur = xmlSecAddChild(keyInfoNode, BAD_CAST "KeyName", NULL);
	if(cur == NULL) {
#ifdef DEBUG_XMLSEC
    	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to create KeyName node\n", 
		func);	
#endif
	    xmlUnlinkNode(keyInfoNode);
	    xmlFreeNode(keyInfoNode);
	    return(NULL);	    
	} 
    }
    if(origin & xmlSecKeyOriginKeyValue) {
	cur = xmlSecAddChild(keyInfoNode, BAD_CAST "KeyValue", NULL);
	if(cur == NULL) {
#ifdef DEBUG_XMLSEC
    	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to create KeyValue  node\n", 
		func);	
#endif
	    xmlUnlinkNode(keyInfoNode);
	    xmlFreeNode(keyInfoNode);
	    return(NULL);	    
	} 
    }
    if(origin & xmlSecKeyOriginX509) {
	cur = xmlSecAddChild(keyInfoNode, BAD_CAST "X509Data", NULL);
	if(cur == NULL) {
#ifdef DEBUG_XMLSEC
    	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to create X509Data  node\n", 
		func);	
#endif
	    xmlUnlinkNode(keyInfoNode);
	    xmlFreeNode(keyInfoNode);
	    return(NULL);	    
	} 
    }    
    if(origin & xmlSecKeyOriginPGP) {
	cur = xmlSecAddChild(keyInfoNode, BAD_CAST "PGPData", NULL);
	if(cur == NULL) {
#ifdef DEBUG_XMLSEC
    	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to create PGPData node\n", 
		func);	
#endif
	    xmlUnlinkNode(keyInfoNode);
	    xmlFreeNode(keyInfoNode);
	    return(NULL);	    
	} 
    }
    
    return(keyInfoNode);
}

/** 
 * xmlDSigObjectNodeCreate:
 * @signatureNode: the Signature node created by xmlDSigSignatureNodeCreate()
 *                  function
 * @id: the node id
 *
 * Adds new Object node to the Signature
 * 
 * Returns pointer to the new Object node or NULL in the case of error 
 */
xmlNodePtr		
xmlDSigObjectNodeCreate(xmlNodePtr signatureNode, const xmlChar *id) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlDSigObjectNodeCreate";
    xmlNodePtr objectNode = NULL;
    
    if(signatureNode == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: Signature node is null\n", 
	    func);	
#endif
	return(NULL);	    
    }

    objectNode = xmlSecAddChild(signatureNode, BAD_CAST "Object", NULL);
    if(objectNode == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s failed to create new Object node\n",
	    func);
#endif
	return(NULL);	
    }
    if(id == NULL) {
	xmlSetProp(objectNode, BAD_CAST "Id", id);
    }

    return(objectNode);
}

/** 
 * xmlDSigManifestNodeCreate
 * @signatureNode: the Signature node created by xmlDSigSignatureNodeCreate()
 *                  function
 * @id: the node id
 *
 * Adds new Manifest node to the Signature
 * 
 * Returns pointer to the new Manifest node or NULL in the case of error 
 */
xmlNodePtr
xmlDSigManifestNodeCreate(xmlNodePtr signatureNode, const xmlChar *id) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlDSigManifestNodeCreate";
    xmlNodePtr objectNode = NULL; 
    xmlNodePtr manifestNode = NULL;
    
    if(signatureNode == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: Signature node is null\n", 
	    func);	
#endif
	return(NULL);	    
    }
    
    objectNode = xmlSecFindChild(signatureNode, BAD_CAST "Object", xmlSecDSigNs);
    if(objectNode == NULL) {
	objectNode = xmlDSigObjectNodeCreate(signatureNode, NULL);
	if(objectNode == NULL) {   
#ifdef DEBUG_XMLSEC
    	    xmlGenericError(xmlGenericErrorContext,
		"%s failed to create new Object node\n",
		func);
#endif
	    return(NULL);	
	}
    }
    
    manifestNode = xmlSecAddChild(objectNode, BAD_CAST "Manifest", NULL);
    if(manifestNode == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s failed to create new Manifest node\n",
	    func);
#endif
	return(NULL);	
    }
    if(id == NULL) {
	xmlSetProp(manifestNode, BAD_CAST "Id", id);
    }
    
    return(manifestNode);
}


