/** 
 * XMLSec library
 *
 * "XML Encryption" implementation
 *  http://www.w3.org/TR/xmlenc-core
 * 
 * See Copyright for the status of this software.
 * 
 * Author: Aleksey Sanin <aleksey@aleksey.com>
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <libxml/tree.h>
#include <libxml/parser.h> 

#include <xmlsec/xmlsec.h>
#include <xmlsec/xmltree.h>
#include <xmlsec/keys.h>
#include <xmlsec/transforms.h>
#include <xmlsec/xmlenc.h>
#include <xmlsec/base64.h>
#include <xmlsec/membuf.h>


typedef struct _xmlEncState {
    xmlEncCtxPtr		ctx;
    xmlSecTransformPtr		first;
    xmlSecTransformPtr		last;
    xmlNodePtr			cipherDataNode;
    int				encrypt;
} xmlEncState, *xmlEncStatePtr;

/** 
 * XML Enc state
 */
static xmlEncStatePtr		xmlEncStateCreate	(xmlEncCtxPtr ctx, 
							 int encrypt);
static void			xmlEncStateDestroy	(xmlEncStatePtr state);
static int			xmlEncStateAddTransform	(xmlEncStatePtr state,
							 xmlSecTransformPtr transform);

#define xmlEncStateGetKeysContext(state) \
    ((( ( state ) != NULL ) && ( ( state )->ctx != NULL ) ) ? \
      (( ( state )->ctx->keysReadContext)) : NULL)


static int			xmlEncEncryptedDataRead	(xmlEncStatePtr state, 
							 xmlNodePtr tmpl,
							 xmlEncResultPtr result);
static int 			xmlEncCipherDataWrite	(xmlNodePtr cipherDataNode,
							 const unsigned char *buf,
							 size_t size);


/**
 * XML Encrypiton context methods
 */
/**
 * xmlEncCtxCreate:
 *
 *
 *
 */
xmlEncCtxPtr		
xmlEncCtxCreate(const xmlSecKeysReadContext *keysReadContext) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncCtxCreate";
    xmlEncCtxPtr ctx;
    
    /*
     * Allocate a new xmlEncCtx and fill the fields.
     */
    ctx = (xmlEncCtxPtr) xmlMalloc(sizeof(xmlEncCtx));
    if(ctx == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: xmlEncCtx malloc failed\n",
	    func);	
#endif 	    
	return(NULL);
    }
    memset(ctx, 0, sizeof(xmlEncCtx));
    
    ctx->keysReadContext = keysReadContext;
    return(ctx);
}

/**
 * xmlEncCtxDestroy
 *
 *
 *
 */
void
xmlEncCtxDestroy(xmlEncCtxPtr ctx) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncCtxDestroy";
    if(ctx == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: context is null\n", 
	    func);	
#endif
	return;	    
    }
    
    memset(ctx, 0, sizeof(xmlEncCtx));
    xmlFree(ctx);
}


/**
 * Encryption
 */
/**
 * xmlEncEncryptMemory
 *
 *
 *
 */
int
xmlEncEncryptMemory(xmlEncCtxPtr ctx, xmlNodePtr tmpl,
		    const unsigned char *buf, size_t size,
		    xmlEncResultPtr *result) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncEncryptMemory";
    xmlEncStatePtr state;
    xmlBufferPtr memBuf;
    xmlEncResultPtr res;
    int ret;
    
    if((ctx == NULL) || (tmpl == NULL) || (buf == NULL)) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: context, tmpl or buff is null\n", 
	    func);	
#endif
	return(-1);	    
    }
    
    if(result != NULL) {
	result = NULL;
    }

    /* first of all, create result and encryption state objects */
    state = xmlEncStateCreate(ctx, 1);
    if(state == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create encryption state\n", 
	    func);	
#endif
	return(-1);	    
    }
    
    res = xmlEncResultCreate(ctx, 1, tmpl);
    if(res == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create result object\n", 
	    func);	
#endif
	xmlEncStateDestroy(state);    
	return(-1);	    		
    }
    
    /*  read and update the Encryption Method and KeyInfo */
    ret = xmlEncEncryptedDataRead(state, tmpl, res);
    if(ret < 0) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to read EncryptedData\n", 
	    func);	
#endif
	xmlEncResultDestroy(res);
	xmlEncStateDestroy(state);    	
	return(-1);	    	
    }
    
    
    /* encrypt the data */
    ret = xmlSecBinTransformWrite(state->first, buf, size);
    if(ret < 0) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to encrypt the buffer\n", 
	    func);	
#endif
	xmlEncResultDestroy(res);
	xmlEncStateDestroy(state);    
	return(-1);	    	
    }
    ret = xmlSecBinTransformFlush(state->first);
    if(ret < 0) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to finalize encryption\n", 
	    func);	
#endif
	xmlEncResultDestroy(res);
	xmlEncStateDestroy(state);    
	return(-1);	    	
    }

    /* update template */
    memBuf = xmlSecMemBufTransformGetBuffer(state->last, 0);
    if(memBuf == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to get memory buffer\n", 
	    func);	
#endif
	xmlEncResultDestroy(res);
	xmlEncStateDestroy(state);    
	return(-1);	    	
    }
    
    ret = xmlEncCipherDataWrite(state->cipherDataNode, 
				xmlBufferContent(memBuf), 
				xmlBufferLength(memBuf));
    if(ret < 0) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to write CipherData node\n", 
	    func);	
#endif
	xmlEncResultDestroy(res);
	xmlEncStateDestroy(state);    
	return(-1);	    	
    }
				
    
    /* TODO: update xmlEncResult */
    
    /* cleanup */
    if(result != NULL) {
	(*result) = res;
    } else {
	xmlEncResultDestroy(res);
    }
    xmlEncStateDestroy(state);    
    return(0);
}

/**
 * xmlEncEncryptUri
 *
 *
 *
 */
int
xmlEncEncryptUri(xmlEncCtxPtr ctx, xmlNodePtr tmpl, const char *uri,
		 xmlEncResultPtr *result) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncEncryptUri";

    if((ctx == NULL) || (tmpl == NULL) || (uri == NULL)) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: context, tmpl or uri is null\n", 
	    func);	
#endif
	return(-1);	    
    }

    /* TODO: */
    return(0);
}

/**
 * xmlEncEncryptXmlNode
 *
 *
 *
 */
int
xmlEncEncryptXmlNode(xmlEncCtxPtr ctx, xmlNodePtr tmpl, xmlNodePtr src,
		 xmlEncResultPtr *result) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncEncryptXmlNode";

    if((ctx == NULL) || (tmpl == NULL) || (src == NULL)) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: context, tmpl or src is null\n", 
	    func);	
#endif
	return(-1);	    
    }


    /* TODO: */
    return(0);
}
							 
/** 
 * XMLS Enc state
 */
/** 
 * xmlEncStateCreate
 *
 *
 *
 */ 
static xmlEncStatePtr
xmlEncStateCreate(xmlEncCtxPtr ctx, int encrypt) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncStateCreate";
    xmlEncStatePtr state;
    
    if(ctx == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: context or tmpl is null\n", 
	    func);	
#endif
	return(NULL);	    
    }
    
    /*
     * Allocate a new xmlEncState and fill the fields.
     */
    state = (xmlEncStatePtr) xmlMalloc(sizeof(xmlEncState));
    if(state == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: xmlEncState malloc failed\n",
	    func);	
#endif 	    
	return(NULL);
    }
    memset(state, 0, sizeof(xmlEncState));
        
    state->encrypt = 1;
    state->ctx = ctx;
    return(state);
}

/** 
 * xmlEncStateDestroy
 *
 *
 *
 */ 
static void
xmlEncStateDestroy(xmlEncStatePtr state) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncStateDestroy";
    if(state == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: context is null\n", 
	    func);	
#endif
	return;	    
    }
    
    if(state->first != NULL) {
	xmlSecBinTransformDestroyAll(state->first);
    } else if(state->last != NULL) {
	xmlSecBinTransformDestroyAll(state->last);
    }

    memset(state, 0, sizeof(xmlEncState));
    xmlFree(state);
}

/**
 * xmlEncStateAddTransform
 *
 *
 *
 *
 */
static int
xmlEncStateAddTransform(xmlEncStatePtr state, xmlSecTransformPtr transform) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncStateAddTransform";

    if((state == NULL) || (transform == NULL)){
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: cipherDataNode is null\n", 
	    func);	
#endif
	return(-1);	    
    }
    if(state->last == NULL) {
	state->first = state->last = transform;
    } else if(xmlSecBinTransformAddAfter(state->last, transform) != NULL) {
	 state->last = transform;
    } else {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to add transform\n", 
	    func);	
#endif
	return(-1);	    
    }
    return(0);
}

/**
 * XML Enc Result
 */
/**
 * xmlEncResultCreate
 *
 *
 *
 *
 */ 		
xmlEncResultPtr		
xmlEncResultCreate(xmlEncCtxPtr ctx, int encrypt, xmlNodePtr node) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncResultCreate";
    xmlEncResultPtr result;
    
    if(ctx == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: context or tmpl is null\n", 
	    func);	
#endif
	return(NULL);	    
    }
    
    /*
     * Allocate a new xmlEncResult and fill the fields.
     */
    result = (xmlEncResultPtr) xmlMalloc(sizeof(xmlEncResult));
    if(result == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: xmlEncResult malloc failed\n",
	    func);	
#endif 	    
	return(NULL);
    }
    memset(result, 0, sizeof(xmlEncResult));
        
    result->ctx = ctx;
    result->self = node;
    result->encrypt = 1;
    return(result);
}

/**
 * xmlEncResultDestoy
 *
 *
 *
 *
 */ 		
void
xmlEncResultDestroy(xmlEncResultPtr result) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncResultDestroy";
    if(result == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: result is null\n", 
	    func);	
#endif
	return;	    
    }
    
    if(result->key != NULL) {
	xmlSecKeyDestroy(result->key, 0);
    }

    memset(result, 0, sizeof(xmlEncResult));
    xmlFree(result);
}



/**
 * xmlEncCipherDataWrite
 *
 *
 *
 *
 */
static int
xmlEncCipherDataWrite(xmlNodePtr cipherDataNode,
		      const unsigned char *buf, size_t size) {
    static const char func[] _UNUSED_VARIABLE_ ="xmlEncCipherDataWrite";

    if(cipherDataNode == NULL){
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: cipherDataNode is null\n", 
	    func);	
#endif
	return(-1);	    
    }
    
    xmlNodeSetContentLen(cipherDataNode, buf, size);
    return(0);
}


/**
 * xmlEncEncryptedDataRead
 *
 *
 */
static int
xmlEncEncryptedDataRead(xmlEncStatePtr state, xmlNodePtr tmpl, xmlEncResultPtr result) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncEncryptedDataRead";
    xmlNodePtr cur;
    xmlNodePtr keyInfoNode = NULL;
    xmlSecTransformPtr encryptionMethod = NULL;
    xmlSecTransformPtr memBuf = NULL;	
    int ret;
    
    if((state == NULL) || (tmpl == NULL) || (result == NULL)) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: state or tmpl node is null\n", 
	    func);	
#endif
	return(-1);	    
    }
    cur = xmlSecGetNextElementNode(tmpl->children);
    
    if(state->encrypt != 1) {
	/* TODO: first transform for decryption is base64 decode */
    }
    
    /* first node is optional EncryptionMethod */
    if((cur != NULL) && (xmlSecCheckNodeName(cur, BAD_CAST "EncryptionMethod", xmlSecEncNs))) {
	encryptionMethod = xmlSecTransformNodeRead(cur, xmlSecUsageEncryptionMethod, 0);
	cur = xmlSecGetNextElementNode(cur->next);
    } else if((state->ctx != NULL) && (state->ctx->encryptionMethod != xmlSecTransformUnknown)){
	/* get encryption method from the context */
	encryptionMethod = xmlSecTransformCreate(state->ctx->encryptionMethod,
						xmlSecUsageEncryptionMethod, 0);
    } else {
#ifdef DEBUG_XMLSEC
	 xmlGenericError(xmlGenericErrorContext,
	    "%s: EncryptionMethod not specified in the tmpl and context\n",
	    func);
#endif		
	return(-1);
    }
    if(encryptionMethod == NULL) {
#ifdef DEBUG_XMLSEC
         xmlGenericError(xmlGenericErrorContext,
    	    "%s: failed to read or create EncryptionMethod\n",
	    func);
#endif		
	return(-1);
    }    
    ret = xmlEncStateAddTransform(state, encryptionMethod);
    if(ret < 0) {    
#ifdef DEBUG_XMLSEC
         xmlGenericError(xmlGenericErrorContext,
    	    "%s: failed to add EncryptionMethod\n",
	    func);
#endif		
	xmlSecTransformDestroy(encryptionMethod, 1); 
	return(-1);
    }
    result->encryptionMethod = encryptionMethod->id;
    
    /* next node is optional KeyInfo */
    if((cur != NULL) && (xmlSecCheckNodeName(cur, BAD_CAST "KeyInfo", xmlSecDSigNs))) {
	keyInfoNode = cur;
	result->key = xmlSecKeyInfoRead(cur, encryptionMethod->id, xmlSecKeyTypePrivate, 
				xmlEncStateGetKeysContext(state));
	cur = xmlSecGetNextElementNode(cur->next);
    }    
    if((result->key == NULL) && (xmlEncStateGetKeysContext(state) != NULL) && 
	      (xmlEncStateGetKeysContext(state)->allowedOrigins & xmlSecKeyOriginKeyManager)) {
	result->key = xmlSecKeySearcCtxFindKey(xmlEncStateGetKeysContext(state),
			encryptionMethod->id, xmlSecKeyTypePrivate,  NULL);
	
    }    
    if(result->key == NULL) {
#ifdef DEBUG_XMLSEC    
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to find encryption key\n",
	    func);
#endif	    
	return(-1);
    }
    ret = xmlSecTransformAddKey(encryptionMethod, result->key);
    if(ret < 0) {
#ifdef DEBUG_XMLSEC    
        xmlGenericError(xmlGenericErrorContext,
    	    "%s: failed to add key\n",
	    func);
#endif	    
	return(-1);
    }
    if((state->encrypt == 1) && (keyInfoNode != NULL)) {
	/* update KeyInfo! */
	ret = xmlSecKeyInfoWrite(keyInfoNode, result->key, xmlSecKeyTypePublic);
	if(ret < 0) {
#ifdef DEBUG_XMLSEC    
	    xmlGenericError(xmlGenericErrorContext,
	        "%s: failed to write \"KeyInfo\"\n",
	        func);
#endif	    
	    return(-1);
	}	
    }

    /* next is required CipherData node */
    if((cur == NULL) || (!xmlSecCheckNodeName(cur, BAD_CAST "CipherData", xmlSecEncNs))) {
#ifdef DEBUG_XMLSEC    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: required element \"CipherData\" missed\n",
	    func);
#endif	    	
	return(-1);
    }
    state->cipherDataNode = cur;
    cur = xmlSecGetNextElementNode(cur->next);

    /* next is optional EncryptionProperties node (we simply ignore it) */
    if((cur != NULL) && (xmlSecCheckNodeName(cur, BAD_CAST "EncryptionProperties", xmlSecDSigNs))) {
	cur = xmlSecGetNextElementNode(cur->next);
    }


    if(state->encrypt == 1) {
	xmlSecTransformPtr base64;
	/* last transform for encryption is base64 encode */
	
	base64 = xmlSecTransformCreate(xmlSecEncBase64Encode, 0, 0);
	if(base64 == NULL) {
#ifdef DEBUG_XMLSEC    
	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to create base64 encode transform\n",
		func);
#endif	    	
	    return(-1);
	}
	ret = xmlEncStateAddTransform(state, base64);
	if(ret < 0) {    
#ifdef DEBUG_XMLSEC
    	    xmlGenericError(xmlGenericErrorContext,
    		"%s: failed to add Base64 encrypt\n",
		func);
#endif		
	    xmlSecTransformDestroy(base64, 1); 
	    return(-1);
	}
    }
    
    /* add mem buf at the end */
    memBuf = xmlSecTransformCreate(xmlSecMemBuf, 0, 0);
    if(memBuf == NULL) {
#ifdef DEBUG_XMLSEC    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create memBuf encode transform\n",
	    func);
#endif	    	
	return(-1);
    }
    ret = xmlEncStateAddTransform(state, memBuf);
    if(ret < 0) {    
#ifdef DEBUG_XMLSEC
	xmlGenericError(xmlGenericErrorContext,
    	    "%s: failed to add Base64 encrypt\n",
	    func);
#endif		
	xmlSecTransformDestroy(memBuf, 1); 
	return(-1);
    }
    
    
    if(cur != NULL) {
#ifdef DEBUG_XMLSEC
	 xmlGenericError(xmlGenericErrorContext,
	    "%s: unexpected node found\n",
	    func);
#endif		
	return(-1);
    }
        
    return(0);
}


