/** 
 * XML Security standards test: XMLEnc
 * 
 * See Copyright for the status of this software.
 * 
 * Author: Aleksey Sanin <aleksey@aleksey.com>
 */
#include <stdlib.h>
#include <string.h>
#include <time.h>


#include <openssl/evp.h>
#include <openssl/rand.h>
#include <openssl/err.h>
#include <openssl/x509.h>
#include <openssl/x509v3.h>

#include <libxml/tree.h>
#include <libxml/parser.h>

#include <xmlsec/xmlsec.h>
#include <xmlsec/keys.h>
#include <xmlsec/keysmngr.h>
#include <xmlsec/xmlenc.h>


void testEncUsage(FILE *f, const char *command);

void testEncUsage(FILE *f, const char *command) {
    fprintf(stderr, "Usage:\n");
    fprintf(f, "%s encrypt|decrypt [<options>] [<filename> [<filename> ... ]]\n", command);
    fprintf(f, "where <options> are\n");
    fprintf(f, "== Key Manager:\n");
    fprintf(f, "  --privkey <file>[,<cafile1>[,<cafile2>[...]]] the private PEM key file\n");
    fprintf(f, "                        and the list of cert file that verify it\n");
    fprintf(f, "  --pubkey <file>       the public PEM key file\n");
    fprintf(f, "== Result generation/printing:\n");
    fprintf(f, "  --dump                dump the debug verification results\n");
    fprintf(f, "== Misc.:\n");
    fprintf(f, "  --retry <number>      repeat operation with each file <number> times\n" );
}

/**
 * Init/Shutdown
 */
int			testEncInit			(int encrypt);
void			testEncShutdown			(void);

/**
 * Parsing the command line
 */
int			testEncReadKeyOrigins		(char *keyOrigins);
int			testEncReadKey			(int privateKey, 
							 char *param);
int 			testEncReadKeysFile		(char *keysFile);
int			testEncReadRetries		(char *retries);







int 			testEnc				(const char* filename, 
							int encrypt);

int 			testEncEncrypt			(xmlDocPtr doc);
int 			testEncDecrypt			(xmlDocPtr doc);


xmlEncCtxPtr encCtx = NULL;
xmlSecSimpleKeyMngrPtr keyMgr = NULL; 
xmlSecKeysReadContext keysReadCtx;
int retryNumber = 1;

clock_t total_time = 0;
int doDump = 0;

int main(int argc, char **argv) {
    int encrypt = 0;    
    int ret = -1;
    int res = -1;
    int pos;
    int argsOk = 0;    

    /*
     * Parse command line
     */
    if( argc < 3 ) {
	fprintf(stderr, "Error: wrong number of arguments.\n");
	goto done;		
    } 
    if(strcmp(argv[1], "encrypt") == 0) {
	encrypt = 1;
    } else if(strcmp(argv[1], "decrypt") == 0) {
	encrypt = 0;
    } else {
        fprintf(stderr, "Error: bad mode \"%s\".\n", argv[1]);
        goto done;
    }
    
    ret = testEncInit(encrypt);
    if(ret < 0) {
	goto done;
    }
    
    /* load options */
    pos = 2;
    while(pos < argc) {
	if(argv[pos][0] != '-') {
	    /* it's filename? */
	    break;
	} else if((strcmp(argv[pos], "--keys") == 0) && (pos + 1 < argc)) {	  
	    ret = testEncReadKeysFile(argv[++pos]);
	} else if((strcmp(argv[pos], "--retry") == 0) && (pos + 1 < argc)) {
	    ret = testEncReadRetries(argv[++pos]);
	} else if((strcmp(argv[pos], "--privkey") == 0) && (pos + 1 < argc)) {
	    ret = testEncReadKey(1, argv[++pos]); 
	} else if((strcmp(argv[pos], "--pubkey") == 0) && (pos + 1 < argc)) {
	    ret = testEncReadKey(0, argv[++pos]); 
	} else if((strcmp(argv[pos], "--keys-origins") == 0) && (pos + 1 < argc)){
	    ret = testEncReadKeyOrigins(argv[++pos]);
	} else if(strcmp(argv[pos], "--dump") == 0) {
	    doDump = 1;
	} else {
	    fprintf(stderr, "Error: bad option \"%s\".\n", argv[pos]);
	    goto done;
	}
	if(ret < 0) {
	    goto done;
	}
	++pos;
    }
    argsOk = 1;

    while(pos < argc) { 
	if(encrypt) {
	    res = testEnc(argv[pos++], 1);	    
	    if(res < 0) {	
		break;
	    }
	} else {
	    res = testEnc(argv[pos++], 0);
	    if(res < 0) {
		break;
	    }
	}
    }
    fprintf(stderr, "Executed %d tests in %ld msec\n", retryNumber, total_time / (CLOCKS_PER_SEC / 1000));

done:
    if(!argsOk) {
	testEncUsage(stderr, argv[0]);
    }
        
    testEncShutdown();
    return((res >= 0) ? 0 : 1);
}

int 
testEnc(const char* filename, int generate) {
    xmlDocPtr doc;
    int ret = 0;
    int i;
    clock_t start_time;
    
    /*
     * build an XML tree from a the file; we need to add default
     * attributes and resolve all character and entities references
     */
    xmlLoadExtDtdDefaultValue = XML_DETECT_IDS | XML_COMPLETE_ATTRS;
    xmlSubstituteEntitiesDefault(1);

    doc = xmlParseFile(filename);
    if (doc == NULL) {
	fprintf(stderr, "Error: unable to parse file \"%s\"\n", filename);
	return(-1);
    }
    
    /*
     * Check the document is of the right kind
     */    
    if(xmlDocGetRootElement(doc) == NULL) {
        fprintf(stderr,"Error: empty document for file \"%s\"\n", filename);
	xmlFreeDoc(doc);
	return(-1);
    }

    start_time = clock();
    for(i = 0; i < retryNumber; ++i) {
        /*
	 * Generate/verify encryptature
	 */      
	if(generate) {
	    fprintf(stderr, "Generating xml encryptature for \"%s\"\n", filename);	
	    ret = testEncEncrypt(doc);
	} else {
	    fprintf(stderr, "Validating xml encryptature in \"%s\"\n", filename);	
	    ret = testEncDecrypt(doc);
	}
    }    
    
    total_time += clock() - start_time;
           
    /*
     * Cleanup
     */ 
    xmlFreeDoc(doc);    

    return(ret);
}
	
int testEncEncrypt(xmlDocPtr doc) {    
    xmlEncResultPtr encResult = NULL;
    xmlChar *result = NULL;	
    int len;
    int ret;
    int res = -1;

    ret = xmlEncEncryptMemory(encCtx, xmlDocGetRootElement(doc), 
			(unsigned char*)"test", 4, &encResult);
    if(ret < 0) {
        fprintf(stderr,"Error: xmlEncGenerate() failed \n");
	goto done;    
    } 
    	
    /*
     * Print document out in default UTF-8 encoding
     */
    xmlDocDumpMemoryEnc(doc, &result, &len, NULL);
    if(result == NULL) {
        fprintf(stderr,"Error: failed to dump document to memory\n");
	goto done;
    }
    fwrite(result, len, 1, stdout);
    res = 0;
    
done:    
    if(result != NULL) {
	xmlFree(result);        
    }
    if(encResult != NULL) {
/*	xmlEncSignatureDestroyAll(encResult); */
    }
    return(res);
}

int testEncDecrypt(xmlDocPtr doc) {    
    xmlEncResultPtr decResult;
    int ret;
    int res = -1;
#if 0    	    
    ret = xmlEncValidate(encCtx, xmlDocGetRootElement(doc), &encryptature);
    if((ret < 0) || (encryptature == NULL)){
	fprintf(stdout,"ERROR\n");
	goto done;
    } 
	    
    if(doDump) {
    	xmlEncDebugSignatureDumpAll(stderr, encryptature);
    }	
	    
    /** 
     * we will simply walk thru and calculate the number of 
     * ok/fails
     */
    memset(&result, 0, sizeof(result));
    testEncGetResult(encryptature, &result, 0);
	
    /* 
     * if one encryptature fails then we are failed. 
     * in real application the rules could be different
     */
    fprintf(stderr, "= Status:\n");
    fprintf(stderr, "== Signatures ok: %d\n", result.encryptaturesOk);
    fprintf(stderr, "== Signatures fail: %d\n", result.encryptaturesFail);
    fprintf(stderr, "== SignedInfo Ref ok: %d\n", result.encryptRefOk);
    fprintf(stderr, "== SignedInfo Ref fail: %d\n", result.encryptRefFail);
    fprintf(stderr, "== Manifest Ref ok: %d\n", result.manifestRefOk);
    fprintf(stderr, "== Manifest Ref fail: %d\n", result.manifestRefFail);

    if(result.encryptaturesFail == 0) {
	fprintf(stdout, "OK\n");  
	res = 0;    
    }else {
	fprintf(stdout, "FAIL\n");
	res = -1;    
    }
    	    	
done:
    if(encryptature != NULL) {
	xmlEncSignatureDestroyAll(encryptature);
    }
#endif     
    return(res);
}

/**
 * Parsing the command line
 */
int			
testEncReadKeyOrigins(char *keyOrigins) {
    xmlSecKeyOrigin res = xmlSecKeyOriginDefault;
    char *p;
    
    p = strtok(keyOrigins, ",");
    while(p != NULL) {
	if(strcmp(p, "keymanager") == 0) {
	    res |= xmlSecKeyOriginKeyManager;
	} else if(strcmp(p, "keyname") == 0) {
	    res |= xmlSecKeyOriginKeyManager | xmlSecKeyOriginKeyName;
	} else if(strcmp(p, "keyvalue") == 0) {
	    res |= xmlSecKeyOriginKeyValue;
	} else if(strcmp(p, "retrieval-doc") == 0) {
	    res |= xmlSecKeyOriginRetrievalDocument;
	} else if(strcmp(p, "retrieval-remote") == 0) {
	    res |= xmlSecKeyOriginRetrievalRemote;
	} else if(strcmp(p, "x509") == 0) {
	    res |= xmlSecKeyOriginX509;
	} else if(strcmp(p, "pgp") == 0) {
	    res |= xmlSecKeyOriginPGP;
	} else {
	    fprintf(stderr, "Error: unknown key origin: \"%s\" (ignored)\n", p);
	    return(-1);
	}
	p = strtok(NULL, ",");
    }    
    keysReadCtx.allowedOrigins = res;
    return(0);
}

int
testEncReadKey	(int privateKey, char *param) {
    char *p;
    xmlSecKeyPtr key;
    int ret;
    
    p = strtok(param, ","); 
    if(privateKey) {
	key = xmlSecSimpleKeyMngrLoadPrivateKey(keyMgr, p, NULL, NULL);
    } else {
	key = xmlSecSimpleKeyMngrLoadPublicKey(keyMgr, p, NULL, NULL);
    }
    if(key == NULL) {
	fprintf(stderr, "Error: failed to load key from \"%s\"\n", p);
	return(-1);
    }
    
    p = strtok(NULL, ",");
    while((p != NULL) && (privateKey)) {
	ret = xmlSecSimpleKeyMngrAddCertToKey(key, p);
	if(ret < 0){
	    fprintf(stderr, "Error: failed to load cert from \"%s\"\n", p);
	    return(-1);
	}
	p = strtok(NULL, ","); 
    }
    /* TODO: */
    return(0);
}

int
testEncReadKeysFile(char *keysFile) {
    int ret;
    
    ret = xmlSecSimpleKeyMngrLoad(keyMgr, keysFile);
    if(ret < 0) {
	fprintf(stderr, "Error: failed to load keys from \"%s\".\n", keysFile);
	return(-1);
    }
    return(0);
}

int
testEncReadRetries(char *retries) {
    
    if(sscanf(retries, "%d", &retryNumber) != 1) {
	fprintf(stderr, "Error: bad number \"%s\".\n", retries);
	return(-1);
    }	    
    return(0);
}

/**
 * Init/Shutdown
 */
int 
testEncInit(int encrypt) {
    int rnd_seed = 0;

    /* 
     * Init OpenSSL
     */
    OpenSSL_add_all_algorithms();
    ERR_load_crypto_strings(); 
    while (RAND_status() != 1) {
	RAND_seed(&rnd_seed, sizeof(rnd_seed));
    }

    /*
     * Init libxml
     */     
    xmlInitParser();
    LIBXML_TEST_VERSION

    /*
     * Init xmlsec
     */
    xmlSecInit();    


    /** 
     * Create Keys managers
     */
    keyMgr = xmlSecSimpleKeyMngrCreate();    
    if(keyMgr == NULL) {
	fprintf(stderr, "Error: failed to create keys manager\n");
	return(-1);
    }
    memset(&keysReadCtx, 0, sizeof(keysReadCtx));

    keysReadCtx.allowedOrigins = xmlSecKeyOriginAll; /* by default all keys are accepted */
    keysReadCtx.maxRetrievals = 1; /* one level retrievals only */
    keysReadCtx.findKeyCallback = xmlSecSimpleKeyMngrFindKey;
    keysReadCtx.findKeyContext = keyMgr;

    /**
     * Init Enc context
     */    
    encCtx = xmlEncCtxCreate(&keysReadCtx);
    if(encCtx == NULL) {
        fprintf(stderr,"Error: failed to create Enc context\n");
	return(-1);
    }
        
    return(0);    
}

void
testEncShutdown(void) {
    /* destroy xmlsec objects */
    if(keyMgr != NULL) {
	xmlSecSimpleKeyMngrDestroy(keyMgr);
    }
    if(encCtx != NULL) {
	xmlEncCtxDestroy(encCtx);
    }

    xmlSecShutdown();
            
    /* 
     * Shutdown libxml
     */
    xmlCleanupParser();
    xmlMemoryDump();

    /**
     * Shutdown OpenSSL
     */    
    RAND_cleanup();
    ERR_free_strings();
    EVP_cleanup();
    X509_TRUST_cleanup();
}

